"""Short-lived deploy ticket storage for Flow Builder beta deploys."""

from __future__ import annotations

import hashlib
import json
import secrets
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from threading import RLock
from typing import Any, Optional


def _utcnow() -> datetime:
    return datetime.now(timezone.utc)


def _normalize_payload(value: Any) -> Any:
    if hasattr(value, "model_dump"):
        return value.model_dump(mode="json", exclude_none=True)
    if isinstance(value, dict):
        return {k: _normalize_payload(v) for k, v in value.items()}
    if isinstance(value, (list, tuple)):
        return [_normalize_payload(v) for v in value]
    return value


def _payload_hash(value: Any) -> str:
    payload = json.dumps(
        _normalize_payload(value),
        sort_keys=True,
        separators=(",", ":"),
    )
    return hashlib.sha256(payload.encode("utf-8")).hexdigest()


@dataclass(frozen=True)
class FlowDeployTicket:
    ticket_id: str
    account_id: str
    flow_api_name: str
    flow_xml: str
    xml_hash: str
    plan_hash: str
    context_hash: str
    audit_hash: str
    audit_requires_acknowledgment: bool
    audit_acknowledged: bool
    issued_at: datetime
    expires_at: datetime


class FlowDeployTicketStore:
    """In-memory beta ticket store. Tickets are short-lived and single-use."""

    _tickets: dict[str, FlowDeployTicket] = {}
    _lock = RLock()
    _ttl_seconds = 15 * 60

    @classmethod
    def issue_ticket(
        cls,
        *,
        account_id: str,
        flow_api_name: str,
        flow_xml: str,
        plan: Any,
        context: Any,
        audit: Any,
        audit_acknowledged: bool,
    ) -> FlowDeployTicket:
        now = _utcnow()
        ticket = FlowDeployTicket(
            ticket_id=secrets.token_urlsafe(24),
            account_id=str(account_id),
            flow_api_name=flow_api_name,
            flow_xml=flow_xml,
            xml_hash=_payload_hash(flow_xml),
            plan_hash=_payload_hash(plan),
            context_hash=_payload_hash(context),
            audit_hash=_payload_hash(audit),
            audit_requires_acknowledgment=bool(
                audit and getattr(audit, "requires_acknowledgment", False)
            ),
            audit_acknowledged=bool(audit_acknowledged),
            issued_at=now,
            expires_at=now + timedelta(seconds=cls._ttl_seconds),
        )
        with cls._lock:
            cls._purge_expired_locked(now)
            cls._tickets[ticket.ticket_id] = ticket
        return ticket

    @classmethod
    def consume_ticket(
        cls,
        ticket_id: str,
        *,
        account_id: str,
    ) -> tuple[Optional[FlowDeployTicket], Optional[str]]:
        now = _utcnow()
        with cls._lock:
            cls._purge_expired_locked(now)
            ticket = cls._tickets.get(ticket_id)
            if ticket is None:
                return None, "missing"
            if ticket.account_id != str(account_id):
                return None, "account_mismatch"
            del cls._tickets[ticket_id]
            return ticket, None

    @classmethod
    def clear_all(cls) -> None:
        with cls._lock:
            cls._tickets.clear()

    @classmethod
    def _purge_expired_locked(cls, now: Optional[datetime] = None) -> None:
        current = now or _utcnow()
        expired = [
            ticket_id
            for ticket_id, ticket in cls._tickets.items()
            if ticket.expires_at <= current
        ]
        for ticket_id in expired:
            cls._tickets.pop(ticket_id, None)



"""FoundryMatch SaaS API helpers for g-gremlin."""

from __future__ import annotations

import os
from typing import Any, Dict, Optional

import click

from g_gremlin.base import (
    AuthenticationError,
    BaseIntegration,
    IntegrationError,
    RateLimitError,
    RetryConfig,
    TransientError,
)
from g_gremlin.credentials import resolve_credentials
from g_gremlin.credentials.provider import CredentialProvider, SecretNotFoundError

DEFAULT_API_BASE = os.getenv("G_GREMLIN_FM_API_BASE") or os.getenv("FOUNDRYOPS_API_URL") or "https://api.foundryops.io"


class FoundryMatchError(RuntimeError):
    """Raised when FoundryMatch SaaS operations fail."""


class FoundryMatchAuthError(FoundryMatchError):
    """Raised when FoundryMatch credentials are invalid."""


class FoundryMatchClient(BaseIntegration):
    """Client for FoundryMatch SaaS enrichment endpoints."""

    SERVICE_NAME = "foundrymatch"

    def __init__(
        self,
        api_base: Optional[str] = None,
        api_key: Optional[str] = None,
        workspace_id: Optional[str] = None,
        credential_provider: Optional[CredentialProvider] = None,
        retry_config: Optional[RetryConfig] = None,
    ) -> None:
        super().__init__(
            credential_provider=credential_provider,
            default_retry_config=retry_config or RetryConfig(),
        )
        creds = None
        if not api_key or not api_base or not workspace_id:
            try:
                creds = resolve_credentials(
                    "foundrymatch",
                    required_keys=["api_base", "api_key", "workspace_id"],
                    credential_provider=credential_provider,
                )
            except SecretNotFoundError:
                creds = None
        self.api_base = (api_base or (creds or {}).get("api_base") or DEFAULT_API_BASE).rstrip("/")
        self.api_key = api_key or (creds or {}).get("api_key")
        self.workspace_id = workspace_id or (creds or {}).get("workspace_id")
        if not self.api_key or not self.workspace_id:
            raise click.ClickException(
                "FoundryMatch SaaS credentials not configured. "
                "Set G_GREMLIN_FM_API_BASE, G_GREMLIN_FM_API_KEY, and G_GREMLIN_FM_WORKSPACE_ID "
                "or run g-gremlin auth set foundrymatch."
            )

    def _headers(self) -> Dict[str, str]:
        return {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }

    def _url(self, path: str) -> str:
        path = path if path.startswith("/") else f"/{path}"
        return f"{self.api_base}{path}"

    def run_enrichment(self, payload: Dict[str, Any]) -> Dict[str, Any]:
        try:
            resp = self._post(self._url("/api/v1/enrichment/run"), headers=self._headers(), json=payload)
            return resp.json()
        except AuthenticationError as exc:
            raise FoundryMatchAuthError("FoundryMatch API key invalid.") from exc
        except (RateLimitError, IntegrationError, TransientError) as exc:
            raise FoundryMatchError(str(exc)) from exc

    def get_job_status(self, job_id: str) -> Dict[str, Any]:
        try:
            resp = self._get(self._url(f"/api/v1/enrichment/run/{job_id}"), headers=self._headers())
            return resp.json()
        except AuthenticationError as exc:
            raise FoundryMatchAuthError("FoundryMatch API key invalid.") from exc
        except (RateLimitError, IntegrationError, TransientError) as exc:
            raise FoundryMatchError(str(exc)) from exc



"""
Amplemarket sales engagement client.

Supports:
- Identity checks
- Sequence discovery
- Contact lookup/listing
- Sequence enrollment (bulk)
"""

from __future__ import annotations

import csv
import logging
import re
import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional
from urllib.parse import parse_qs, quote, urlparse

import click

from g_gremlin.base import (
    AuthenticationError,
    BaseIntegration,
    IntegrationError,
    RateLimitError,
    RetryConfig,
    TransientError,
)
from g_gremlin.contracts import BatchOutcome, CapabilityCheck
from g_gremlin.credentials import resolve_credentials
from g_gremlin.credentials.provider import CredentialProvider, SecretNotFoundError
from g_gremlin.pacing import AmplemarketPacingProfile, get_amplemarket_profile

logger = logging.getLogger(__name__)

API_BASE = "https://api.amplemarket.com"
DEFAULT_RETRY_CONFIG = RetryConfig(max_attempts=4, base_delay=1.2, max_delay=15.0)
ERROR_STATUSES = (400, 402, 409, 422)
CONTACT_FIELDS = {
    "email",
    "first_name",
    "last_name",
    "title",
    "company_name",
    "linkedin_url",
    "phone",
}

RATE_LIMITS = {
    "people_search": 300,  # per minute
    "people_find": 350,  # per minute
    "sequence_leads": 30,  # per minute
    "default": 500,  # per minute
}

LEAD_LIST_BATCH_MAX = 10000
LEAD_LIST_MAX = 60000
LEAD_LIST_TYPES = {"linkedin", "email", "titles_and_company"}
LEAD_LIST_TYPE_ALIASES = {
    "title_and_company": "titles_and_company",
}

SEQUENCE_DATA_FIELD_RE = re.compile(r"^[a-z][a-z0-9_]{0,254}$")
SEQUENCE_DATA_FIELDS_MAX = 50
SEQUENCE_DATA_VALUE_MAX = 1024
SEQUENCE_LEADS_MAX = 250


def _prune_payload(value: Any) -> Any:
    if isinstance(value, dict):
        cleaned: Dict[str, Any] = {}
        for key, item in value.items():
            pruned = _prune_payload(item)
            if pruned is None:
                continue
            if isinstance(pruned, (dict, list)) and not pruned:
                continue
            if pruned == "":
                continue
            cleaned[key] = pruned
        return cleaned
    if isinstance(value, list):
        cleaned_list = []
        for item in value:
            pruned = _prune_payload(item)
            if pruned is None:
                continue
            if isinstance(pruned, (dict, list)) and not pruned:
                continue
            if pruned == "":
                continue
            cleaned_list.append(pruned)
        return cleaned_list
    return value


class _RateLimiter:
    def __init__(self, rate_per_minute: float) -> None:
        self.capacity = max(rate_per_minute, 0.0)
        self.tokens = self.capacity
        self.refill_rate = self.capacity / 60.0 if self.capacity else 0.0
        self.last_check = time.monotonic()

    def acquire(self, sleep_fn, tokens: float = 1.0) -> None:
        if not self.refill_rate:
            return
        now = time.monotonic()
        elapsed = max(now - self.last_check, 0.0)
        self.tokens = min(self.capacity, self.tokens + elapsed * self.refill_rate)
        self.last_check = now
        if self.tokens >= tokens:
            self.tokens -= tokens
            return
        needed = tokens - self.tokens
        wait_for = needed / self.refill_rate if self.refill_rate else 0.0
        if wait_for > 0:
            sleep_fn(wait_for)
        self.last_check = time.monotonic()
        self.tokens = max(self.tokens - tokens, 0.0)


class AmplemarketError(RuntimeError):
    """Base Amplemarket client error."""


class AmplemarketAuthError(AmplemarketError, AuthenticationError):
    """Authentication failed."""

    def __init__(self, message: str):
        AuthenticationError.__init__(self, message)


class AmplemarketRateLimitError(AmplemarketError, RateLimitError):
    """Raised when Amplemarket rate limits requests."""

    def __init__(self, message: str, retry_after: float | None = None):
        RateLimitError.__init__(self, message, retry_after=retry_after)


class InsufficientCreditsError(AmplemarketError):
    """Raised when the API reports insufficient credits."""


class MissingFieldError(AmplemarketError):
    """Raised when a required field is missing."""


class UnsupportedValueError(AmplemarketError):
    """Raised when an unsupported value is provided."""


@dataclass
class AmplemarketContact:
    email: Optional[str] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    title: Optional[str] = None
    company_name: Optional[str] = None
    linkedin_url: Optional[str] = None
    phone: Optional[str] = None

    def to_api_format(self) -> Dict[str, Any]:
        payload: Dict[str, Any] = {}
        if self.email:
            payload["email"] = self.email
        if self.first_name:
            payload["first_name"] = self.first_name
        if self.last_name:
            payload["last_name"] = self.last_name
        if self.title:
            payload["title"] = self.title
        if self.company_name:
            payload["company_name"] = self.company_name
        if self.linkedin_url:
            payload["linkedin_url"] = self.linkedin_url
        if self.phone:
            payload["phone"] = self.phone
        return payload


@dataclass
class AmplemarketSequence:
    id: str
    name: str
    status: str = ""
    step_count: int = 0


@dataclass
class AmplemarketLeadList:
    id: str
    name: str
    status: str = ""
    lead_type: str = ""
    shared: Optional[bool] = None
    visible: Optional[bool] = None
    owner_email: Optional[str] = None
    total_leads: Optional[int] = None
    raw: Dict[str, Any] = field(default_factory=dict)


@dataclass
class AmplemarketEnrichmentRequest:
    id: str
    status: str = ""
    request_type: str = ""
    total_requested: Optional[int] = None
    total_enriched: Optional[int] = None
    raw: Dict[str, Any] = field(default_factory=dict)


@dataclass
class AmplemarketSearchPage:
    results: List[Dict[str, Any]]
    pagination: Dict[str, Any] = field(default_factory=dict)
    raw: Dict[str, Any] = field(default_factory=dict)


@dataclass
class AmplemarketSequenceLeadPayload:
    email: Optional[str] = None
    linkedin_url: Optional[str] = None
    data: Dict[str, Any] = field(default_factory=dict)
    overrides: Dict[str, Any] = field(default_factory=dict)

    def to_api_format(self) -> Dict[str, Any]:
        payload: Dict[str, Any] = {}
        if self.email:
            payload["email"] = self.email
        if self.linkedin_url:
            payload["linkedin_url"] = self.linkedin_url
        if self.data:
            payload["data"] = self.data
        if self.overrides:
            payload["overrides"] = self.overrides
        return payload


@dataclass
class EnrollmentResult:
    total: int
    total_added: int
    duplicate_emails: List[str]
    duplicate_linkedin_urls: List[str]
    already_in_sequence: List[str]
    excluded: List[str]
    added_ids_by_email: Dict[str, str] = field(default_factory=dict)
    added_ids_by_linkedin: Dict[str, str] = field(default_factory=dict)

    def provider_id_for(
        self,
        email: Optional[str] = None,
        linkedin_url: Optional[str] = None,
    ) -> Optional[str]:
        email_key = (email or "").strip().lower()
        if email_key:
            provider_id = self.added_ids_by_email.get(email_key)
            if provider_id:
                return provider_id
        linkedin_key = (linkedin_url or "").strip().lower()
        if linkedin_key:
            provider_id = self.added_ids_by_linkedin.get(linkedin_key)
            if provider_id:
                return provider_id
        return None

    def to_batch_outcome(self) -> BatchOutcome:
        outcome = BatchOutcome()
        for _ in range(max(self.total_added, 0)):
            outcome.succeeded.append({"status": "added"})
        for email in self.excluded:
            outcome.failed.append({"email": email, "status": "excluded"})
        for email in self.duplicate_emails:
            outcome.retryable_failed.append({"email": email, "status": "duplicate"})
        for url in self.duplicate_linkedin_urls:
            outcome.retryable_failed.append({"linkedin_url": url, "status": "duplicate"})
        return outcome


def get_amplemarket_api_key(credential_provider: CredentialProvider | None = None) -> str:
    provider = credential_provider or CredentialProvider()
    try:
        creds = resolve_credentials("amplemarket", required_keys=["api_key"], credential_provider=provider)
        return creds["api_key"]
    except SecretNotFoundError as exc:
        raise click.ClickException(
            "Amplemarket not configured. Set up with:\n"
            "  g-gremlin auth set amplemarket                 # Store in keychain\n"
            "  export G_GREMLIN_AMPLEMARKET_API_KEY=your-key  # Or use env var"
        ) from exc


class AmplemarketClient(BaseIntegration):
    SERVICE_NAME = "amplemarket"

    def __init__(
        self,
        api_key: Optional[str] = None,
        credential_provider: CredentialProvider | None = None,
        retry_config: RetryConfig | None = None,
        api_base: Optional[str] = None,
    ) -> None:
        super().__init__(credential_provider=credential_provider, default_retry_config=retry_config or DEFAULT_RETRY_CONFIG)
        self.api_key = api_key or get_amplemarket_api_key(credential_provider)
        self.api_base = api_base or API_BASE
        self._rate_limiters: Dict[str, _RateLimiter] = {}

    def _headers(self) -> Dict[str, str]:
        return {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }

    def _normalize_path(self, path: str) -> str:
        parsed = urlparse(path) if path.startswith("http") else urlparse(path)
        return parsed.path or path.split("?")[0]

    def _rate_limit_key(self, path: str) -> str:
        normalized = self._normalize_path(path)
        if normalized.startswith("/people/search"):
            return "people_search"
        if normalized.startswith("/people/find"):
            return "people_find"
        if "/sequences/" in normalized and normalized.endswith("/leads"):
            return "sequence_leads"
        return "default"

    def _rate_limit(self, path: str) -> None:
        key = self._rate_limit_key(path)
        limit = RATE_LIMITS.get(key, RATE_LIMITS["default"])
        limiter = self._rate_limiters.get(key)
        if limiter is None:
            limiter = _RateLimiter(limit)
            self._rate_limiters[key] = limiter
        limiter.acquire(self._sleep)

    def _retry_after_seconds(self, response) -> float | None:
        retry_after = super()._retry_after_seconds(response)
        if retry_after is None:
            return None
        return retry_after + 1.0

    def _am_request(
        self,
        method: str,
        path: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        allow_statuses: tuple[int, ...] = (),
    ):
        url = path if path.startswith("http") else f"{self.api_base}{path}"
        merged_allow = tuple(set(allow_statuses).union(ERROR_STATUSES))
        self._rate_limit(path)
        try:
            resp = self._request(
                method,
                url,
                params=params,
                json=json,
                headers=self._headers(),
                allow_statuses=merged_allow,
            )
        except AuthenticationError as exc:
            raise AmplemarketAuthError(str(exc)) from exc
        except RateLimitError as exc:
            raise AmplemarketRateLimitError(str(exc), retry_after=exc.retry_after) from exc
        except (IntegrationError, TransientError) as exc:
            raise AmplemarketError(str(exc)) from exc

        if resp.ok or resp.status_code in allow_statuses:
            return resp

        self._raise_for_errors(resp)

    def _raise_for_errors(self, resp) -> None:
        detail = f"Amplemarket error {resp.status_code}"
        try:
            payload = resp.json()
        except Exception:
            raise AmplemarketError(f"{detail}: {resp.text[:200]}")

        code = ""
        message = ""
        if isinstance(payload, dict):
            errors = payload.get("_errors") or payload.get("errors")
            if isinstance(errors, list) and errors:
                err = errors[0] or {}
                code = str(err.get("code") or "").strip()
                message = err.get("detail") or err.get("message") or err.get("title") or ""

        if code == "insufficient_credits":
            raise InsufficientCreditsError(message or detail)
        if code == "missing_field":
            raise MissingFieldError(message or detail)
        if code == "unsupported_value":
            raise UnsupportedValueError(message or detail)
        if code:
            raise AmplemarketError(message or detail)

        if isinstance(payload, dict):
            message = payload.get("message") or payload.get("detail") or ""
        raise AmplemarketError(message or detail)

    def _parse_sequence(self, data: Dict[str, Any]) -> Optional[AmplemarketSequence]:
        if not isinstance(data, dict):
            return None
        sequence_id = data.get("id") or data.get("sequence_id") or data.get("sequenceId")
        if sequence_id is None:
            return None
        name = data.get("name") or data.get("sequence_name") or data.get("sequenceName") or ""
        status = data.get("status") or data.get("state") or ""
        step_count = data.get("step_count") or data.get("stepCount")
        if step_count is None:
            steps = data.get("steps")
            if isinstance(steps, list):
                step_count = len(steps)
        try:
            step_count_int = int(step_count or 0)
        except (TypeError, ValueError):
            step_count_int = 0

        return AmplemarketSequence(
            id=str(sequence_id),
            name=str(name),
            status=str(status),
            step_count=step_count_int,
        )

    def _parse_contact(self, data: Dict[str, Any]) -> Optional[AmplemarketContact]:
        if not isinstance(data, dict):
            return None
        return AmplemarketContact(
            email=data.get("email") or data.get("email_address") or data.get("emailAddress"),
            first_name=data.get("first_name") or data.get("firstName"),
            last_name=data.get("last_name") or data.get("lastName"),
            title=data.get("title"),
            company_name=data.get("company_name") or data.get("company") or data.get("companyName"),
            linkedin_url=data.get("linkedin_url") or data.get("linkedinUrl"),
            phone=data.get("phone") or data.get("phone_number") or data.get("phoneNumber"),
        )

    def _coerce_contact(self, lead: AmplemarketContact | Dict[str, Any]) -> AmplemarketContact:
        if isinstance(lead, AmplemarketContact):
            return lead
        if isinstance(lead, dict):
            filtered = {key: lead.get(key) for key in CONTACT_FIELDS if key in lead}
            return AmplemarketContact(**filtered)
        raise AmplemarketError("Invalid lead payload provided.")

    def _sanitize_sequence_data(self, data: Dict[str, Any]) -> Dict[str, str]:
        if not isinstance(data, dict):
            raise UnsupportedValueError("Sequence data must be a dictionary.")
        if len(data) > SEQUENCE_DATA_FIELDS_MAX:
            raise UnsupportedValueError("Sequence data supports up to 50 fields.")
        cleaned: Dict[str, str] = {}
        for key, value in data.items():
            if key is None:
                continue
            key_str = str(key).strip()
            if not key_str or not SEQUENCE_DATA_FIELD_RE.match(key_str):
                raise UnsupportedValueError(f"Invalid data field name: {key_str}")
            if value is None:
                continue
            if isinstance(value, (dict, list, tuple, set)):
                raise UnsupportedValueError(f"Invalid data field value for '{key_str}'")
            value_str = str(value)
            if len(value_str) > SEQUENCE_DATA_VALUE_MAX:
                raise UnsupportedValueError(f"Data field '{key_str}' exceeds 1024 characters.")
            cleaned[key_str] = value_str
        return cleaned

    def _coerce_sequence_lead(self, lead: Any) -> Dict[str, Any]:
        if isinstance(lead, AmplemarketSequenceLeadPayload):
            payload = lead.to_api_format()
        elif isinstance(lead, AmplemarketContact):
            payload = lead.to_api_format()
        elif isinstance(lead, dict):
            payload = {k: v for k, v in lead.items() if v is not None}
        else:
            raise AmplemarketError("Invalid lead payload provided.")

        if "data" in payload:
            payload["data"] = self._sanitize_sequence_data(payload.get("data") or {})

        if not payload.get("email") and not payload.get("linkedin_url"):
            raise MissingFieldError("Lead must include email or linkedin_url.")

        return payload

    def _extract_provider_id(self, item: Dict[str, Any]) -> Optional[str]:
        for key in (
            "id",
            "lead_id",
            "leadId",
            "person_id",
            "personId",
            "contact_id",
            "contactId",
        ):
            value = item.get(key)
            if value is None:
                continue
            text = str(value).strip()
            if text:
                return text
        return None

    def _extract_added_ids(
        self,
        payload: Dict[str, Any],
        data: Dict[str, Any],
    ) -> tuple[Dict[str, str], Dict[str, str]]:
        added_ids_by_email: Dict[str, str] = {}
        added_ids_by_linkedin: Dict[str, str] = {}

        def ingest(items: Any) -> None:
            if not isinstance(items, list):
                return
            for item in items:
                if not isinstance(item, dict):
                    continue
                provider_id = self._extract_provider_id(item)
                if not provider_id:
                    continue
                email = str(
                    item.get("email")
                    or item.get("email_address")
                    or item.get("lead_email")
                    or ""
                ).strip().lower()
                linkedin = str(
                    item.get("linkedin_url")
                    or item.get("linkedinUrl")
                    or item.get("linkedin")
                    or item.get("linkedin_profile")
                    or ""
                ).strip().lower()
                if email:
                    added_ids_by_email[email] = provider_id
                if linkedin:
                    added_ids_by_linkedin[linkedin] = provider_id

        # Common list payload variants used by sequence add endpoints.
        for container in (data, payload):
            if not isinstance(container, dict):
                continue
            for key in (
                "added_leads",
                "added_contacts",
                "added_people",
                "people_added",
                "contacts_added",
                "leads_added",
                "added",
                "results",
            ):
                ingest(container.get(key))

            map_by_email = container.get("added_by_email")
            if isinstance(map_by_email, dict):
                for email, provider_id in map_by_email.items():
                    email_key = str(email or "").strip().lower()
                    provider_id_text = str(provider_id or "").strip()
                    if email_key and provider_id_text:
                        added_ids_by_email[email_key] = provider_id_text

            map_by_linkedin = (
                container.get("added_by_linkedin")
                or container.get("added_by_linkedin_url")
            )
            if isinstance(map_by_linkedin, dict):
                for linkedin, provider_id in map_by_linkedin.items():
                    linkedin_key = str(linkedin or "").strip().lower()
                    provider_id_text = str(provider_id or "").strip()
                    if linkedin_key and provider_id_text:
                        added_ids_by_linkedin[linkedin_key] = provider_id_text

        return added_ids_by_email, added_ids_by_linkedin

    def _parse_enrollment_result(self, payload: Dict[str, Any]) -> EnrollmentResult:
        data = payload.get("data") if isinstance(payload.get("data"), dict) else payload
        if not isinstance(data, dict):
            data = {}

        skipped = data.get("skipped") if isinstance(data.get("skipped"), dict) else {}

        total = int(data.get("total") or data.get("total_leads") or 0)
        total_added = int(
            data.get("total_added")
            or data.get("total_added_to_sequence")
            or data.get("added")
            or data.get("added_count")
            or 0
        )
        duplicate_emails = list(
            data.get("duplicate_emails")
            or data.get("duplicates")
            or skipped.get("duplicate_emails")
            or []
        )
        duplicate_linkedin_urls = list(
            data.get("duplicate_linkedin_urls")
            or skipped.get("duplicate_linkedin_urls")
            or []
        )
        already_in_sequence = list(
            data.get("already_in_sequence")
            or data.get("already_in_sequence_emails")
            or data.get("already_in_sequence_and_skipped")
            or skipped.get("already_in_sequence")
            or []
        )
        excluded = list(
            data.get("excluded")
            or data.get("excluded_emails")
            or data.get("in_exclusion_list_and_skipped")
            or skipped.get("excluded")
            or []
        )
        added_ids_by_email, added_ids_by_linkedin = self._extract_added_ids(payload, data)

        return EnrollmentResult(
            total=total,
            total_added=total_added,
            duplicate_emails=[str(x) for x in duplicate_emails if x],
            duplicate_linkedin_urls=[str(x) for x in duplicate_linkedin_urls if x],
            already_in_sequence=[str(x) for x in already_in_sequence if x],
            excluded=[str(x) for x in excluded if x],
            added_ids_by_email=added_ids_by_email,
            added_ids_by_linkedin=added_ids_by_linkedin,
        )

    def _split_href(self, href: str) -> tuple[str, Dict[str, Any]]:
        parsed = urlparse(href)
        path = parsed.path or href
        params: Dict[str, Any] = {}
        if parsed.query:
            parsed_params = parse_qs(parsed.query)
            for key, value in parsed_params.items():
                if not value:
                    continue
                params[key] = value[0] if len(value) == 1 else value
        if path and not path.startswith("/") and not path.startswith("http"):
            path = f"/{path}"
        return path, params

    def _follow_pages(
        self,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        page_size: Optional[int] = None,
        data_keys: tuple[str, ...] = ("data", "results"),
    ) -> Iterable[Dict[str, Any]]:
        next_path = path
        next_params = dict(params or {})
        if page_size:
            next_params.setdefault("page[size]", page_size)

        while next_path:
            resp = self._am_request("GET", next_path, params=next_params)
            payload = resp.json() if resp.content else {}
            data = []
            if isinstance(payload, dict):
                for key in data_keys:
                    candidate = payload.get(key)
                    if candidate is not None:
                        data = candidate
                        break
            if isinstance(data, list):
                for item in data:
                    yield item

            links = payload.get("_links") or payload.get("links") or {}
            next_link = links.get("next") or {}
            href = next_link.get("href") or next_link.get("url")
            if not href:
                break
            next_path, next_params = self._split_href(href)

    # ------------------------------------------------------------------
    # Public API methods
    # ------------------------------------------------------------------
    def whoami(self) -> Dict[str, Any]:
        resp = self._am_request("GET", "/account-info")
        payload = resp.json() if resp.content else {}
        return payload if isinstance(payload, dict) else {"raw": payload}

    def probe_capabilities(self) -> List[CapabilityCheck]:
        checks = [
            self._probe_endpoint("account", "/account-info"),
            self._probe_endpoint("sequences", "/sequences"),
            self._probe_endpoint("contacts", "/contacts", params={"page[size]": 1}),
        ]
        return checks

    def _probe_endpoint(
        self,
        name: str,
        path: str,
        *,
        params: Optional[Dict[str, Any]] = None,
    ) -> CapabilityCheck:
        try:
            resp = self._am_request("GET", path, params=params, allow_statuses=(403,))
            if resp.status_code == 403:
                detail = f"{resp.status_code} {resp.reason}".strip()
                return CapabilityCheck(name=name, status="denied", detail=detail)
            return CapabilityCheck(name=name, status="ok", detail="")
        except AmplemarketError as exc:
            return CapabilityCheck(name=name, status="error", detail=str(exc))

    def list_sequences(self, page_size: int = 50) -> List[AmplemarketSequence]:
        sequences: List[AmplemarketSequence] = []
        for item in self._follow_pages("/sequences", page_size=page_size):
            seq = self._parse_sequence(item)
            if seq:
                sequences.append(seq)
        return sequences

    def get_contact(self, contact_id: str) -> Optional[AmplemarketContact]:
        resp = self._am_request("GET", f"/contacts/{contact_id}", allow_statuses=(404,))
        if resp.status_code == 404:
            return None
        payload = resp.json() if resp.content else {}
        data = payload.get("data") if isinstance(payload, dict) else payload
        return self._parse_contact(data) if isinstance(data, dict) else None

    def find_contact_by_email(self, email: str) -> Optional[AmplemarketContact]:
        email_safe = quote(email)
        resp = self._am_request("GET", f"/contacts/email/{email_safe}", allow_statuses=(404,))
        if resp.status_code == 404:
            return None
        payload = resp.json() if resp.content else {}
        data = payload.get("data") if isinstance(payload, dict) else payload
        return self._parse_contact(data) if isinstance(data, dict) else None

    def list_contacts(self, limit: int = 100, page_size: int = 100) -> List[AmplemarketContact]:
        contacts: List[AmplemarketContact] = []
        for item in self._follow_pages("/contacts", page_size=page_size):
            contact = self._parse_contact(item)
            if contact:
                contacts.append(contact)
                if limit and len(contacts) >= limit:
                    break
        return contacts

    def _parse_enrichment_request(self, payload: Dict[str, Any]) -> AmplemarketEnrichmentRequest:
        data = payload.get("data") if isinstance(payload.get("data"), dict) else payload
        if not isinstance(data, dict):
            data = {}
        request_id = str(data.get("id") or data.get("request_id") or "")
        status = str(data.get("status") or data.get("state") or "")
        request_type = str(data.get("type") or data.get("request_type") or "")
        total_requested = data.get("total_requested") or data.get("total") or data.get("total_leads")
        total_enriched = data.get("total_enriched") or data.get("enriched") or data.get("total_enriched_leads")
        try:
            total_requested = int(total_requested) if total_requested is not None else None
        except (TypeError, ValueError):
            total_requested = None
        try:
            total_enriched = int(total_enriched) if total_enriched is not None else None
        except (TypeError, ValueError):
            total_enriched = None
        return AmplemarketEnrichmentRequest(
            id=request_id,
            status=status,
            request_type=request_type,
            total_requested=total_requested,
            total_enriched=total_enriched,
            raw=payload if isinstance(payload, dict) else {},
        )

    def _parse_lead_list(self, payload: Dict[str, Any]) -> Optional[AmplemarketLeadList]:
        if not isinstance(payload, dict):
            return None
        list_id = payload.get("id") or payload.get("lead_list_id") or payload.get("leadListId")
        if list_id is None:
            return None
        name = payload.get("name") or payload.get("lead_list_name") or payload.get("leadListName") or ""
        status = payload.get("status") or payload.get("state") or ""
        lead_type = payload.get("type") or payload.get("lead_type") or payload.get("leadType") or ""
        owner = payload.get("owner") if isinstance(payload.get("owner"), dict) else {}
        owner_email = owner.get("email") if owner else None
        total_leads = payload.get("total") or payload.get("total_leads") or payload.get("count")
        try:
            total_leads = int(total_leads) if total_leads is not None else None
        except (TypeError, ValueError):
            total_leads = None
        return AmplemarketLeadList(
            id=str(list_id),
            name=str(name),
            status=str(status),
            lead_type=str(lead_type),
            shared=payload.get("shared"),
            visible=payload.get("visible"),
            owner_email=owner_email,
            total_leads=total_leads,
            raw=payload,
        )

    def search_people(self, request: Dict[str, Any]) -> AmplemarketSearchPage:
        payload = _prune_payload(dict(request or {}))
        resp = self._am_request("POST", "/people/search", json=payload)
        result = resp.json() if resp.content else {}
        if not isinstance(result, dict):
            return AmplemarketSearchPage(results=[], pagination={}, raw={"raw": result})
        results = result.get("data") or result.get("people") or result.get("results") or []
        if not isinstance(results, list):
            results = []
        pagination = result.get("_pagination") or result.get("pagination") or {}
        return AmplemarketSearchPage(results=results, pagination=pagination or {}, raw=result)

    def find_person(self, **params: Any) -> Optional[Dict[str, Any]]:
        cleaned = _prune_payload(params)
        resp = self._am_request("GET", "/people/find", params=cleaned, allow_statuses=(404,))
        if resp.status_code == 404:
            return None
        payload = resp.json() if resp.content else {}
        return payload if isinstance(payload, dict) else {"raw": payload}

    def start_person_enrichment_batch(
        self,
        leads: List[Dict[str, Any]],
        *,
        reveal_email: Optional[bool] = None,
        reveal_phone_numbers: Optional[bool] = None,
    ) -> AmplemarketEnrichmentRequest:
        payload = _prune_payload(
            {
                "leads": leads,
                "reveal_email": reveal_email,
                "reveal_phone_numbers": reveal_phone_numbers,
            }
        )
        resp = self._am_request("POST", "/people/enrichment-requests", json=payload)
        data = resp.json() if resp.content else {}
        return self._parse_enrichment_request(data if isinstance(data, dict) else {})

    def get_person_enrichment_batch(
        self,
        request_id: str,
        *,
        page_size: Optional[int] = None,
        page_after: Optional[str] = None,
        page_before: Optional[str] = None,
    ) -> Dict[str, Any]:
        params: Dict[str, Any] = {}
        if page_size:
            params["page[size]"] = page_size
        if page_after:
            params["page[after]"] = page_after
        if page_before:
            params["page[before]"] = page_before
        resp = self._am_request("GET", f"/people/enrichment-requests/{request_id}", params=params)
        payload = resp.json() if resp.content else {}
        return payload if isinstance(payload, dict) else {"raw": payload}

    def cancel_person_enrichment_batch(self, request_id: str) -> Dict[str, Any]:
        resp = self._am_request(
            "PATCH",
            f"/people/enrichment-requests/{request_id}",
            json={"status": "canceled"},
        )
        payload = resp.json() if resp.content else {}
        return payload if isinstance(payload, dict) else {"raw": payload}

    def search_companies(self, request: Dict[str, Any]) -> AmplemarketSearchPage:
        """
        Search companies via POST /companies/search.

        Note: This endpoint is documented in guides but may change without notice.
        """
        payload = _prune_payload(dict(request or {}))
        resp = self._am_request("POST", "/companies/search", json=payload)
        result = resp.json() if resp.content else {}
        if not isinstance(result, dict):
            return AmplemarketSearchPage(results=[], pagination={}, raw={"raw": result})
        results = result.get("data") or result.get("companies") or result.get("results") or []
        if not isinstance(results, list):
            results = []
        pagination = result.get("_pagination") or result.get("pagination") or {}
        return AmplemarketSearchPage(results=results, pagination=pagination or {}, raw=result)

    def find_company(self, **params: Any) -> Optional[Dict[str, Any]]:
        cleaned = _prune_payload(params)
        resp = self._am_request("GET", "/companies/find", params=cleaned, allow_statuses=(404,))
        if resp.status_code == 404:
            return None
        payload = resp.json() if resp.content else {}
        return payload if isinstance(payload, dict) else {"raw": payload}

    def start_company_enrichment_batch(
        self,
        leads: List[Dict[str, Any]],
    ) -> AmplemarketEnrichmentRequest:
        payload = _prune_payload({"companies": leads})
        resp = self._am_request("POST", "/companies/enrichment-requests", json=payload)
        data = resp.json() if resp.content else {}
        return self._parse_enrichment_request(data if isinstance(data, dict) else {})

    def get_company_enrichment_batch(
        self,
        request_id: str,
        *,
        page_size: Optional[int] = None,
        page_after: Optional[str] = None,
        page_before: Optional[str] = None,
    ) -> Dict[str, Any]:
        params: Dict[str, Any] = {}
        if page_size:
            params["page[size]"] = page_size
        if page_after:
            params["page[after]"] = page_after
        if page_before:
            params["page[before]"] = page_before
        resp = self._am_request("GET", f"/companies/enrichment-requests/{request_id}", params=params)
        payload = resp.json() if resp.content else {}
        return payload if isinstance(payload, dict) else {"raw": payload}

    def cancel_company_enrichment_batch(self, request_id: str) -> Dict[str, Any]:
        resp = self._am_request(
            "PATCH",
            f"/companies/enrichment-requests/{request_id}",
            json={"status": "canceled"},
        )
        payload = resp.json() if resp.content else {}
        return payload if isinstance(payload, dict) else {"raw": payload}

    def list_lead_lists(self, limit: int = 100, page_size: int = 50) -> List[AmplemarketLeadList]:
        lead_lists: List[AmplemarketLeadList] = []
        for item in self._follow_pages("/lead-lists", page_size=page_size, data_keys=("lead_lists", "data", "results")):
            parsed = self._parse_lead_list(item)
            if parsed:
                lead_lists.append(parsed)
                if limit and len(lead_lists) >= limit:
                    break
        return lead_lists

    def create_lead_list(
        self,
        *,
        name: str,
        lead_type: str,
        leads: List[Dict[str, Any]],
        owner_email: Optional[str] = None,
        shared: Optional[bool] = None,
        visible: Optional[bool] = None,
    ) -> AmplemarketLeadList:
        normalized_type = LEAD_LIST_TYPE_ALIASES.get(lead_type, lead_type)
        normalized_type = str(normalized_type or "").strip().lower()
        if normalized_type not in LEAD_LIST_TYPES:
            raise UnsupportedValueError(f"Unsupported lead list type: {lead_type}")
        self._validate_lead_list_entries(leads, normalized_type)
        payload = _prune_payload(
            {
                "name": name,
                "type": normalized_type,
                "owner": {"email": owner_email} if owner_email else None,
                "shared": shared,
                "visible": visible,
                "leads": leads,
            }
        )
        resp = self._am_request("POST", "/lead-lists", json=payload)
        data = resp.json() if resp.content else {}
        parsed = self._parse_lead_list(data if isinstance(data, dict) else {})
        if parsed:
            return parsed
        raise AmplemarketError("Unexpected response while creating lead list.")

    def get_lead_list(
        self,
        list_id: str,
        *,
        page_size: Optional[int] = None,
        page_after: Optional[str] = None,
        page_before: Optional[str] = None,
    ) -> Dict[str, Any]:
        params: Dict[str, Any] = {}
        if page_size:
            params["page[size]"] = page_size
        if page_after:
            params["page[after]"] = page_after
        if page_before:
            params["page[before]"] = page_before
        resp = self._am_request("GET", f"/lead-lists/{list_id}", params=params)
        payload = resp.json() if resp.content else {}
        return payload if isinstance(payload, dict) else {"raw": payload}

    def add_leads_to_list(
        self,
        list_id: str,
        leads: List[Dict[str, Any]],
        *,
        lead_type: str,
        chunk_size: int = LEAD_LIST_BATCH_MAX,
        max_total: int = LEAD_LIST_MAX,
    ) -> Dict[str, Any]:
        normalized_type = LEAD_LIST_TYPE_ALIASES.get(lead_type, lead_type)
        normalized_type = str(normalized_type or "").strip().lower()
        if normalized_type not in LEAD_LIST_TYPES:
            raise UnsupportedValueError(f"Unsupported lead list type: {lead_type}")
        if len(leads) > max_total:
            raise AmplemarketError(f"Lead list supports up to {max_total} leads.")
        self._validate_lead_list_entries(leads, normalized_type)
        chunk_size = min(max(chunk_size, 1), LEAD_LIST_BATCH_MAX)

        def _chunk(items: List[Dict[str, Any]], size: int) -> Iterable[List[Dict[str, Any]]]:
            for i in range(0, len(items), size):
                yield items[i : i + size]

        responses: List[Dict[str, Any]] = []
        total_added = 0
        total_submitted = 0
        chunks = 0
        for chunk in _chunk(leads, chunk_size):
            payload = _prune_payload(
                {
                    "type": normalized_type,
                    "leads": chunk,
                }
            )
            resp = self._am_request("POST", f"/lead-lists/{list_id}/leads", json=payload)
            data = resp.json() if resp.content else {}
            if isinstance(data, dict):
                responses.append(data)
                total_submitted += int(data.get("total") or len(chunk))
                total_added += int(data.get("total_added_to_lead_list") or data.get("total_added") or 0)
            else:
                responses.append({"raw": data})
                total_submitted += len(chunk)
            chunks += 1

        return {
            "list_id": list_id,
            "lead_type": normalized_type,
            "total_submitted": total_submitted,
            "total_added": total_added,
            "chunks": chunks,
            "responses": responses,
        }

    def _extract_status(self, payload: Dict[str, Any]) -> str:
        if not isinstance(payload, dict):
            return ""
        if isinstance(payload.get("data"), dict):
            status = payload["data"].get("status") or payload["data"].get("state")
            if status:
                return str(status)
        status = payload.get("status") or payload.get("state") or payload.get("request_status")
        return str(status) if status else ""

    def _validate_lead_list_entries(self, leads: List[Dict[str, Any]], lead_type: str) -> None:
        invalid_indices: List[int] = []
        for idx, lead in enumerate(leads):
            if not isinstance(lead, dict):
                invalid_indices.append(idx + 1)
                continue
            if lead_type == "linkedin":
                if not lead.get("linkedin_url"):
                    invalid_indices.append(idx + 1)
            elif lead_type == "email":
                if not lead.get("email"):
                    invalid_indices.append(idx + 1)
            elif lead_type == "titles_and_company":
                if not lead.get("title") or not (lead.get("company_name") or lead.get("company_domain")):
                    invalid_indices.append(idx + 1)

        if invalid_indices:
            sample = ", ".join(str(i) for i in invalid_indices[:10])
            raise MissingFieldError(
                f"{len(invalid_indices)} invalid leads detected (rows: {sample}). "
                "Check required fields for the lead list type."
            )

    def wait_for_enrichment(
        self,
        path: str,
        *,
        poll_interval: float = 30.0,
        max_wait: Optional[float] = None,
    ) -> Dict[str, Any]:
        start = time.monotonic()
        while True:
            resp = self._am_request("GET", path)
            payload = resp.json() if resp.content else {}
            if not isinstance(payload, dict):
                return {"raw": payload}
            status = self._extract_status(payload).lower()
            if status and status in {"completed", "failed", "canceled", "cancelled", "done", "finished"}:
                return payload
            if status and "complete" in status:
                return payload
            if max_wait is not None and (time.monotonic() - start) >= max_wait:
                return payload
            retry_after = self._retry_after_seconds(resp)
            delay = retry_after if retry_after is not None else poll_interval
            if delay > 0:
                self._sleep(delay)

    def add_to_sequence(
        self,
        sequence_id: str,
        leads: List[AmplemarketContact | AmplemarketSequenceLeadPayload | Dict[str, Any]],
        *,
        overrides: Optional[Dict[str, Any]] = None,
        settings: Optional[Dict[str, Any]] = None,
        leads_distribution: Optional[str] = None,
        mailboxes: Optional[List[str]] = None,
        scheduler_start_time: Optional[str] = None,
        ignore_recently_contacted: Optional[bool] = None,
        ignore_exclusion_list: Optional[bool] = None,
        ignore_duplicate_leads_in_other_draft_sequences: Optional[bool] = None,
        ignore_duplicate_leads_in_other_active_sequences: Optional[bool] = None,
        skip_email_validation: Optional[bool] = None,
        skip_email_checks: Optional[bool] = None,
        skip_phone_checks: Optional[bool] = None,
    ) -> EnrollmentResult:
        payload_leads: List[Dict[str, Any]] = []
        for lead in leads:
            payload = self._coerce_sequence_lead(lead)
            if payload:
                payload_leads.append(payload)
        if not payload_leads:
            raise AmplemarketError("No valid leads provided for enrollment.")
        if len(payload_leads) > SEQUENCE_LEADS_MAX:
            raise AmplemarketError(f"Sequence enrollment supports up to {SEQUENCE_LEADS_MAX} leads per request.")

        overrides_payload: Dict[str, Any] = dict(overrides or {})
        if ignore_recently_contacted is not None:
            overrides_payload["ignore_recently_contacted"] = ignore_recently_contacted
        if ignore_exclusion_list is not None:
            overrides_payload["ignore_exclusion_list"] = ignore_exclusion_list
        if ignore_duplicate_leads_in_other_draft_sequences is not None:
            overrides_payload["ignore_duplicate_leads_in_other_draft_sequences"] = (
                ignore_duplicate_leads_in_other_draft_sequences
            )
        if ignore_duplicate_leads_in_other_active_sequences is not None:
            overrides_payload["ignore_duplicate_leads_in_other_active_sequences"] = (
                ignore_duplicate_leads_in_other_active_sequences
            )

        settings_payload: Dict[str, Any] = dict(settings or {})
        if leads_distribution is not None:
            settings_payload["leads_distribution"] = leads_distribution
        if mailboxes:
            settings_payload["mailboxes"] = mailboxes
        if scheduler_start_time is not None:
            settings_payload["scheduler_start_time"] = scheduler_start_time
        if skip_email_validation is not None:
            settings_payload["skip_email_validation"] = skip_email_validation
        if skip_email_checks is not None:
            settings_payload["skip_email_checks"] = skip_email_checks
        if skip_phone_checks is not None:
            settings_payload["skip_phone_checks"] = skip_phone_checks

        resp = self._am_request(
            "POST",
            f"/sequences/{sequence_id}/leads",
            json=_prune_payload(
                {
                    "leads": payload_leads,
                    "overrides": overrides_payload,
                    "settings": settings_payload,
                }
            ),
        )
        payload = resp.json() if resp.content else {}
        return self._parse_enrollment_result(payload if isinstance(payload, dict) else {})

    def add_to_sequence_bulk(
        self,
        leads: List[AmplemarketContact | AmplemarketSequenceLeadPayload | Dict[str, Any]],
        sequence_id: str,
        pacing_profile: AmplemarketPacingProfile | None = None,
        *,
        overrides: Optional[Dict[str, Any]] = None,
        settings: Optional[Dict[str, Any]] = None,
        leads_distribution: Optional[str] = None,
        mailboxes: Optional[List[str]] = None,
        scheduler_start_time: Optional[str] = None,
        ignore_recently_contacted: Optional[bool] = None,
        ignore_exclusion_list: Optional[bool] = None,
        ignore_duplicate_leads_in_other_draft_sequences: Optional[bool] = None,
        ignore_duplicate_leads_in_other_active_sequences: Optional[bool] = None,
        skip_email_validation: Optional[bool] = None,
        skip_email_checks: Optional[bool] = None,
        skip_phone_checks: Optional[bool] = None,
    ) -> BatchOutcome:
        outcome = BatchOutcome()
        profile = pacing_profile or get_amplemarket_profile("safe")
        downshifted = False
        leads_list: List[AmplemarketContact | Dict[str, Any]] = list(leads)

        idx = 0
        total_leads = len(leads_list)
        while idx < total_leads:
            batch_size_limit = min(profile.leads_per_request, SEQUENCE_LEADS_MAX)
            batch = leads_list[idx : idx + batch_size_limit]
            batch_size = len(batch)
            if batch_size == 0:
                break
            batch_payloads: List[Dict[str, Any]] = []
            batch_identifiers: List[Dict[str, Optional[str]]] = []

            for lead in batch:
                try:
                    payload = self._coerce_sequence_lead(lead)
                except AmplemarketError as exc:
                    outcome.failed.append({"email": None, "error": str(exc)})
                    continue
                batch_payloads.append(payload)
                batch_identifiers.append(
                    {
                        "email": payload.get("email"),
                        "linkedin_url": payload.get("linkedin_url"),
                    }
                )

            if not batch_payloads:
                idx += batch_size
                continue

            try:
                result = self.add_to_sequence(
                    sequence_id,
                    batch_payloads,
                    overrides=overrides,
                    settings=settings,
                    leads_distribution=leads_distribution,
                    mailboxes=mailboxes,
                    scheduler_start_time=scheduler_start_time,
                    ignore_recently_contacted=ignore_recently_contacted,
                    ignore_exclusion_list=ignore_exclusion_list,
                    ignore_duplicate_leads_in_other_draft_sequences=ignore_duplicate_leads_in_other_draft_sequences,
                    ignore_duplicate_leads_in_other_active_sequences=ignore_duplicate_leads_in_other_active_sequences,
                    skip_email_validation=skip_email_validation,
                    skip_email_checks=skip_email_checks,
                    skip_phone_checks=skip_phone_checks,
                )
            except AmplemarketRateLimitError as exc:
                if profile.name == "turbo" and not downshifted:
                    logger.warning("Amplemarket rate limited; downshifting to safe pacing.")
                    profile = get_amplemarket_profile("safe")
                    downshifted = True
                for identifier in batch_identifiers:
                    outcome.retryable_failed.append(
                        {
                            "email": identifier.get("email"),
                            "linkedin_url": identifier.get("linkedin_url"),
                            "error": str(exc),
                            "status": "rate_limited",
                        }
                    )
                if profile.inter_call_delay > 0:
                    self._sleep(profile.inter_call_delay)
                idx += batch_size
                continue
            duplicate_set = set(result.duplicate_emails) | set(result.duplicate_linkedin_urls)
            excluded_set = set(result.excluded) | set(result.already_in_sequence)

            for identifier in batch_identifiers:
                email = identifier.get("email")
                linkedin = identifier.get("linkedin_url")
                if (email and email in duplicate_set) or (linkedin and linkedin in duplicate_set):
                    outcome.retryable_failed.append(
                        {"email": email, "linkedin_url": linkedin, "status": "duplicate"}
                    )
                    continue
                if (email and email in excluded_set) or (linkedin and linkedin in excluded_set):
                    outcome.failed.append({"email": email, "linkedin_url": linkedin, "status": "excluded"})
                    continue
                outcome.succeeded.append({"email": email, "linkedin_url": linkedin, "status": "added"})

            if profile.inter_call_delay > 0:
                self._sleep(profile.inter_call_delay)
            idx += batch_size

        return outcome


def _load_contacts_from_csv(csv_path: str, email_col: str) -> List[AmplemarketContact]:
    path = Path(csv_path)
    if not path.exists():
        raise AmplemarketError(f"CSV file not found: {csv_path}")

    contacts: List[AmplemarketContact] = []
    with path.open(newline="", encoding="utf-8") as handle:
        reader = csv.DictReader(handle)
        for row in reader:
            email = (row.get(email_col) or "").strip()
            if not email:
                continue
            contacts.append(
                AmplemarketContact(
                    email=email,
                    first_name=(row.get("first_name") or "").strip() or None,
                    last_name=(row.get("last_name") or "").strip() or None,
                    title=(row.get("title") or "").strip() or None,
                    company_name=(row.get("company_name") or "").strip() or None,
                    linkedin_url=(row.get("linkedin_url") or "").strip() or None,
                    phone=(row.get("phone") or "").strip() or None,
                )
            )

    if not contacts:
        raise AmplemarketError("No valid contacts found in CSV.")

    return contacts


def _load_sequence_leads_from_csv(
    csv_path: str,
    email_col: str = "email",
    linkedin_col: Optional[str] = "linkedin_url",
    data_columns: Optional[List[str]] = None,
) -> List[Dict[str, Any]]:
    path = Path(csv_path)
    if not path.exists():
        raise AmplemarketError(f"CSV file not found: {csv_path}")

    data_columns = data_columns or []
    leads: List[Dict[str, Any]] = []
    with path.open(newline="", encoding="utf-8") as handle:
        reader = csv.DictReader(handle)
        for row in reader:
            email = (row.get(email_col) or "").strip()
            linkedin_url = (row.get(linkedin_col) or "").strip() if linkedin_col else ""
            lead: Dict[str, Any] = {
                "email": email or None,
                "linkedin_url": linkedin_url or None,
                "first_name": (row.get("first_name") or "").strip() or None,
                "last_name": (row.get("last_name") or "").strip() or None,
                "title": (row.get("title") or "").strip() or None,
                "company_name": (row.get("company_name") or "").strip() or None,
                "phone": (row.get("phone") or "").strip() or None,
            }
            data: Dict[str, Any] = {}
            for col in data_columns:
                value = (row.get(col) or "").strip()
                if value:
                    data[col] = value
            if data:
                lead["data"] = data
            lead = _prune_payload(lead)
            if not lead.get("email") and not lead.get("linkedin_url"):
                continue
            leads.append(lead)

    if not leads:
        raise AmplemarketError("No valid leads found in CSV.")
    return leads


def enroll_from_csv(
    csv_path: str,
    sequence_id: str,
    email_col: str = "email",
    linkedin_col: Optional[str] = None,
    data_columns: Optional[List[str]] = None,
    pacing_profile: AmplemarketPacingProfile | None = None,
) -> BatchOutcome:
    leads = _load_sequence_leads_from_csv(
        csv_path=csv_path,
        email_col=email_col,
        linkedin_col=linkedin_col,
        data_columns=data_columns,
    )
    client = AmplemarketClient()
    return client.add_to_sequence_bulk(
        leads,
        sequence_id=sequence_id,
        pacing_profile=pacing_profile,
    )


__all__ = [
    "AmplemarketError",
    "AmplemarketAuthError",
    "AmplemarketRateLimitError",
    "InsufficientCreditsError",
    "MissingFieldError",
    "UnsupportedValueError",
    "AmplemarketContact",
    "AmplemarketSequence",
    "AmplemarketLeadList",
    "AmplemarketEnrichmentRequest",
    "AmplemarketSearchPage",
    "AmplemarketSequenceLeadPayload",
    "EnrollmentResult",
    "AmplemarketClient",
    "get_amplemarket_api_key",
    "enroll_from_csv",
]



"""Assign global sales regions from ISO2 country codes."""

from __future__ import annotations

from typing import Any, Dict, Optional

# Core region sets
NA_CODES = {"US", "CA", "MX"}

LATAM_CODES = {
    "AG",
    "AI",
    "AR",
    "AW",
    "BB",
    "BL",
    "BM",
    "BO",
    "BQ",
    "BR",
    "BS",
    "BZ",
    "CL",
    "CO",
    "CR",
    "CU",
    "CW",
    "DM",
    "DO",
    "EC",
    "FK",
    "GD",
    "GF",
    "GP",
    "GT",
    "GY",
    "HN",
    "HT",
    "JM",
    "KN",
    "KY",
    "LC",
    "MF",
    "MQ",
    "MS",
    "NI",
    "PA",
    "PE",
    "PM",
    "PR",
    "PY",
    "SR",
    "SV",
    "SX",
    "TC",
    "TT",
    "UY",
    "VC",
    "VE",
    "VG",
    "VI",
}

APJ_CODES = {
    "AF",
    "AS",
    "AU",
    "BD",
    "BN",
    "BT",
    "CC",
    "CK",
    "CN",
    "CX",
    "FJ",
    "FM",
    "GU",
    "HK",
    "ID",
    "IN",
    "JP",
    "KG",
    "KH",
    "KI",
    "KP",
    "KR",
    "KZ",
    "LA",
    "LK",
    "MH",
    "MM",
    "MN",
    "MO",
    "MP",
    "MV",
    "MY",
    "NC",
    "NF",
    "NP",
    "NR",
    "NU",
    "NZ",
    "PF",
    "PG",
    "PH",
    "PK",
    "PN",
    "PW",
    "SB",
    "SG",
    "TH",
    "TJ",
    "TK",
    "TL",
    "TM",
    "TO",
    "TV",
    "TW",
    "UM",
    "UZ",
    "VN",
    "VU",
    "WF",
    "WS",
}

EMEA_CODES = {
    "AD",
    "AE",
    "AL",
    "AM",
    "AO",
    "AT",
    "AX",
    "AZ",
    "BA",
    "BE",
    "BF",
    "BG",
    "BH",
    "BI",
    "BJ",
    "BW",
    "BY",
    "CD",
    "CF",
    "CG",
    "CH",
    "CI",
    "CM",
    "CV",
    "CY",
    "CZ",
    "DE",
    "DJ",
    "DK",
    "DZ",
    "EE",
    "EG",
    "EH",
    "ER",
    "ES",
    "ET",
    "FI",
    "FO",
    "FR",
    "GA",
    "GB",
    "GE",
    "GG",
    "GH",
    "GI",
    "GL",
    "GM",
    "GN",
    "GQ",
    "GR",
    "GW",
    "HR",
    "HU",
    "IE",
    "IL",
    "IM",
    "IO",
    "IQ",
    "IR",
    "IS",
    "IT",
    "JE",
    "JO",
    "KE",
    "KM",
    "KW",
    "LB",
    "LI",
    "LR",
    "LS",
    "LT",
    "LU",
    "LV",
    "LY",
    "MA",
    "MC",
    "MD",
    "ME",
    "MG",
    "MK",
    "ML",
    "MR",
    "MT",
    "MU",
    "MW",
    "MZ",
    "NA",
    "NE",
    "NG",
    "NL",
    "NO",
    "OM",
    "PL",
    "PS",
    "PT",
    "QA",
    "RE",
    "RO",
    "RS",
    "RU",
    "RW",
    "SA",
    "SC",
    "SD",
    "SE",
    "SH",
    "SI",
    "SJ",
    "SK",
    "SL",
    "SM",
    "SN",
    "SO",
    "SS",
    "ST",
    "SY",
    "SZ",
    "TD",
    "TG",
    "TN",
    "TR",
    "TZ",
    "UA",
    "UG",
    "VA",
    "XK",
    "YE",
    "YT",
    "ZA",
    "ZM",
    "ZW",
}

# Synonyms frequently seen in CRM exports
SYNONYMS = {
    "UNITED STATES": "US",
    "UNITED STATES OF AMERICA": "US",
    "USA": "US",
    "UNITED KINGDOM": "GB",
    "GREAT BRITAIN": "GB",
    "BRITAIN": "GB",
    "UK": "GB",
    "EN": "GB",
    "ENG": "GB",
    "SCOTLAND": "GB",
    "WALES": "GB",
    "UAE": "AE",
    "KSA": "SA",
    "SOUTH KOREA": "KR",
    "NORTH KOREA": "KP",
    "VIET NAM": "VN",
    "SAUDI ARABIA": "SA",
    "UNITED ARAB EMIRATES": "AE",
    "SOUTH AFRICA": "ZA",
    "MEXICO": "MX",
    "CANADA": "CA",
    "AUSTRALIA": "AU",
    "NEW ZEALAND": "NZ",
}

REGION_BY_CODE = {code: "NA" for code in NA_CODES}
REGION_BY_CODE.update({code: "LATAM" for code in LATAM_CODES})
REGION_BY_CODE.update({code: "EMEA" for code in EMEA_CODES})
REGION_BY_CODE.update({code: "APJ" for code in APJ_CODES})


def _normalize_country_code(value: Optional[str]) -> str:
    if not value:
        return ""
    candidates = str(value).strip().upper()
    if not candidates:
        return ""
    return SYNONYMS.get(candidates, candidates)


def _assign_global_region(code: str) -> str:
    return REGION_BY_CODE.get(code, "")


def assign_global_region(value: str | None) -> Dict[str, Any]:
    """Return the FoundryMatch global region for a 2-letter ISO country code."""

    code = _normalize_country_code(value)
    if not code:
        return {"value": "", "meta": {"reason": "empty", "input": value}}
    if len(code) != 2:
        return {"value": "", "meta": {"reason": "invalid_code", "input": value}}

    region = _assign_global_region(code)
    if not region:
        return {"value": "", "meta": {"reason": "unknown_code", "code": code}}

    return {"value": region, "meta": {"code": code, "region": region}}



"""Stable JSON CLI contracts for IDE-agent consumption."""

from __future__ import annotations

from pathlib import Path
from typing import Any

from g_gremlin.autopilot.ai_provider import AIProviderStore
from g_gremlin.autopilot.audit_log import AuditLog
from g_gremlin.autopilot.concierge_executor import ConciergeExecutor
from g_gremlin.autopilot.concierge_planner import ConciergePlanner
from g_gremlin.autopilot.discovery import DiscoveryRunner, DiscoveryStore
from g_gremlin.autopilot.field_provenance import FieldProvenanceStore
from g_gremlin.autopilot.identity_graph import IdentityGraph
from g_gremlin.autopilot.modules import ModuleStore
from g_gremlin.autopilot.scout_config_store import ScoutConfigStore
from g_gremlin.scout.engine import ScoutEngine
from g_gremlin.scout.loader import ScoutLoader


def cli_discovery_run(
    state_dir: Path,
    *,
    crm: str = "salesforce",
    org_alias: str | None = None,
    user_credentials: set[str] | None = None,
) -> dict[str, Any]:
    """Run discovery and return the report as a JSON-serializable dict."""
    creds = user_credentials or set()
    store = DiscoveryStore(state_dir)
    loader = ScoutLoader(user_credentials=creds)
    engine = ScoutEngine(loader, user_credentials=creds)
    overrides = ScoutConfigStore(state_dir).load_all()
    ai_store = AIProviderStore(state_dir)
    ai_config = ai_store.load_config()
    runner = DiscoveryRunner(
        store,
        engine,
        crm=crm,
        org_alias=org_alias,
        overrides=overrides,
        ai_config=ai_config if ai_config and ai_config.enabled else None,
    )
    report = runner.run()
    return {"status": "completed", "run_id": report.run_id, "report": report.to_dict()}


def cli_discovery_latest(state_dir: Path) -> dict[str, Any]:
    """Return the latest discovery report."""
    store = DiscoveryStore(state_dir)
    run_id = store.latest_run_id()
    if not run_id:
        return {"status": "empty", "run": None, "report": None}
    run = store.load_run(run_id)
    report = store.load_report(run_id)
    return {
        "status": "ok",
        "run": run.to_dict() if run else None,
        "report": report.to_dict() if report else None,
    }


def cli_concierge_preview(
    state_dir: Path,
    *,
    run_id: str | None = None,
    user_credentials: set[str] | None = None,
) -> dict[str, Any]:
    """Preview concierge plan for the given (or latest) discovery run."""
    creds = user_credentials or set()
    store = DiscoveryStore(state_dir)
    target_run_id = run_id or store.latest_run_id()
    if not target_run_id:
        return {"plan": None, "report_run_id": None}
    report = store.load_report(target_run_id)
    if not report:
        return {"plan": None, "report_run_id": None}

    loader = ScoutLoader(user_credentials=creds)
    engine = ScoutEngine(loader, user_credentials=creds)
    planner = ConciergePlanner(
        report,
        ScoutConfigStore(state_dir),
        ModuleStore(state_dir),
        loader,
        engine,
    )
    plan = planner.plan()
    return {"plan": plan.model_dump(exclude_none=False), "report_run_id": report.run_id}


def cli_concierge_apply(
    state_dir: Path,
    *,
    run_id: str,
    action_ids: list[str] | None = None,
    dry_run: bool = True,
    actor: str = "cli",
    user_credentials: set[str] | None = None,
) -> dict[str, Any]:
    """Apply concierge plan and return the receipt."""
    creds = user_credentials or set()
    store = DiscoveryStore(state_dir)
    report = store.load_report(run_id)
    if not report:
        return {"error": f"Discovery run not found: {run_id}"}

    loader = ScoutLoader(user_credentials=creds)
    engine = ScoutEngine(loader, user_credentials=creds)
    planner = ConciergePlanner(
        report,
        ScoutConfigStore(state_dir),
        ModuleStore(state_dir),
        loader,
        engine,
    )
    plan = planner.plan()
    if action_ids is not None:
        selected = [action for action in plan.actions if action.action_id in set(action_ids)]
        plan = plan.model_copy(
            update={
                "actions": selected,
                "stats": ConciergePlanner.compute_stats(selected),
            }
        )

    audit = AuditLog(Path(state_dir) / "db" / "platform.sqlite")
    executor = ConciergeExecutor(
        ScoutConfigStore(state_dir),
        ModuleStore(state_dir),
        state_dir,
        audit_log=audit,
    )
    receipt = executor.execute(plan, dry_run=dry_run, actor=actor)
    return {"receipt": receipt.model_dump(exclude_none=False), "report_run_id": report.run_id}


def cli_substrate_stats(state_dir: Path) -> dict[str, Any]:
    """Return identity/provenance aggregate stats for reporting."""
    db_path = Path(state_dir) / "db" / "platform.sqlite"
    identity = IdentityGraph(db_path)
    provenance = FieldProvenanceStore(db_path)
    return {
        "identity": identity.stats(),
        "provenance": {
            **provenance.stats(),
            "by_source_system": provenance.source_system_counts(),
        },
    }


def cli_substrate_identity_lookup(
    state_dir: Path,
    *,
    source_system: str | None = None,
    source_record_id: str | None = None,
    source_org_id: str | None = None,
    email: str | None = None,
    domain: str | None = None,
) -> dict[str, Any]:
    """Resolve a canonical entity and linked records using stable lookup keys."""
    normalized_system = str(source_system or "").strip().lower()
    normalized_record_id = str(source_record_id or "").strip()
    normalized_org_id = str(source_org_id or "").strip() or None
    normalized_email = str(email or "").strip()
    normalized_domain = str(domain or "").strip()

    has_source_lookup = bool(normalized_system or normalized_record_id)
    if has_source_lookup and not (normalized_system and normalized_record_id):
        return {"error": "source_system and source_record_id are required together"}
    if not has_source_lookup and not normalized_email and not normalized_domain:
        return {"error": "provide source lookup, email, or domain"}

    db_path = Path(state_dir) / "db" / "platform.sqlite"
    identity = IdentityGraph(db_path)
    entity: dict[str, Any] | None = None
    lookup_kind = "unknown"
    if normalized_system and normalized_record_id:
        entity = identity.find_entity_by_record(
            normalized_system,
            normalized_record_id,
            normalized_org_id,
        )
        lookup_kind = "source_record"
    elif normalized_email:
        entity = identity.find_entity_by_email(normalized_email)
        lookup_kind = "email"
    else:
        entity = identity.find_entity_by_domain(normalized_domain)
        lookup_kind = "domain"

    if not entity:
        return {"found": False, "lookup_kind": lookup_kind, "entity": None, "linked_records": []}

    entity_id = str(entity.get("entity_id") or "")
    linked_records = identity.get_linked_records(entity_id)
    return {
        "found": True,
        "lookup_kind": lookup_kind,
        "entity": entity,
        "linked_records": linked_records,
    }


def cli_substrate_provenance_entity(
    state_dir: Path,
    *,
    entity_id: str,
    field_name: str | None = None,
    include_drift: bool = True,
    limit: int = 200,
) -> dict[str, Any]:
    """Return observations + authoritative values for a canonical entity."""
    normalized_entity_id = str(entity_id or "").strip()
    normalized_field_name = str(field_name or "").strip().lower() or None
    if not normalized_entity_id:
        return {"error": "entity_id is required"}

    db_path = Path(state_dir) / "db" / "platform.sqlite"
    provenance = FieldProvenanceStore(db_path)
    observations = provenance.list_entity_observations(
        normalized_entity_id,
        field_name=normalized_field_name,
        limit=limit,
    )

    if normalized_field_name:
        authoritative = provenance.get_authoritative_value(normalized_entity_id, normalized_field_name)
        drift = provenance.detect_drift(normalized_entity_id, normalized_field_name) if include_drift else []
        return {
            "entity_id": normalized_entity_id,
            "field_name": normalized_field_name,
            "observations": observations,
            "authoritative": authoritative,
            "drift": drift,
        }

    field_names: list[str] = []
    seen_fields: set[str] = set()
    for observation in observations:
        name = str(observation.get("field_name") or "").strip().lower()
        if not name or name in seen_fields:
            continue
        seen_fields.add(name)
        field_names.append(name)

    authoritative_by_field: dict[str, dict[str, Any]] = {}
    drift_by_field: dict[str, list[dict[str, Any]]] = {}
    for name in field_names:
        authoritative = provenance.get_authoritative_value(normalized_entity_id, name)
        if authoritative is not None:
            authoritative_by_field[name] = authoritative
        if include_drift:
            drift_rows = provenance.detect_drift(normalized_entity_id, name)
            if drift_rows:
                drift_by_field[name] = drift_rows

    return {
        "entity_id": normalized_entity_id,
        "observations": observations,
        "authoritative_by_field": authoritative_by_field,
        "drift_by_field": drift_by_field if include_drift else {},
    }


def cli_substrate_stale_fields(
    state_dir: Path,
    *,
    older_than_days: int = 90,
    limit: int = 200,
) -> dict[str, Any]:
    """Return stale provenance rows for cleanup/reporting."""
    db_path = Path(state_dir) / "db" / "platform.sqlite"
    provenance = FieldProvenanceStore(db_path)
    rows = provenance.stale_fields(older_than_days=older_than_days)
    max_limit = max(1, min(int(limit), 1000))
    results = rows[:max_limit]
    return {
        "older_than_days": max(int(older_than_days), 0),
        "count": len(results),
        "truncated": len(rows) > max_limit,
        "results": results,
    }


__all__ = [
    "cli_discovery_run",
    "cli_discovery_latest",
    "cli_concierge_preview",
    "cli_concierge_apply",
    "cli_substrate_stats",
    "cli_substrate_identity_lookup",
    "cli_substrate_provenance_entity",
    "cli_substrate_stale_fields",
]



from __future__ import annotations

from typing import Any, Dict

from .normalize_phone_to_e164 import normalize_phone_to_e164


def fix_phone_format(text: str, *, country: str = "US") -> Dict[str, Any]:
    result = normalize_phone_to_e164(text, country=country or "US")
    if not result.get("valid"):
        result["value"] = ""
    meta = dict(result.get("meta") or {})
    meta["deprecated_alias_for"] = "normalize_phone_to_e164"
    result["meta"] = meta
    return result



"""
Basic smoke tests for remaining operators.

These tests ensure operators don't crash and return expected structure.
Not comprehensive, but catches basic regressions.
"""

import pytest
import sys
import os

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "..", "src"))

# Disable external dependencies for smoke tests
os.environ["FM_COMPANY_TO_DOMAIN_REGISTRY"] = "false"
os.environ["FM_COMPANY_TO_DOMAIN_WIKIDATA"] = "false"
os.environ["FM_COMPANY_TO_DOMAIN_DNS"] = "false"
os.environ["FM_COMPANY_TO_DOMAIN_LLM"] = "false"
os.environ["TRANSFORM_ENABLE_LLM_FALLBACK"] = "false"

from fmatch.saas.operators.normalize_company import normalize_company
from fmatch.saas.operators.company_to_domain import company_to_domain
from fmatch.saas.operators.email_domain_match import email_domain_match
from fmatch.saas.operators.extract_domain_root import extract_domain_root
from fmatch.saas.operators.extract_subdomain import extract_subdomain
from fmatch.saas.operators.parse_email_pattern import parse_email_pattern
from fmatch.saas.operators.fix_phone_format import fix_phone_format
from fmatch.saas.operators.normalize_job_function import normalize_job_function
from fmatch.saas.operators.classify_revenue_band import classify_revenue_band
from fmatch.saas.operators.calculate_lead_score import calculate_lead_score
from fmatch.saas.operators.assign_territory import assign_territory
from fmatch.saas.operators.state_name_to_code import state_name_to_code


class TestNormalizeCompany:
    """Basic tests for normalize_company."""

    @pytest.mark.parametrize("input_company,should_clean", [
        ("Acme Corporation", True),
        ("Acme Corp", True),
        ("Acme Inc.", True),
        ("Acme LLC", True),
        ("Acme, Inc.", True),
    ])
    def test_removes_legal_suffixes(self, input_company, should_clean):
        """Test that legal suffixes are removed."""
        result = normalize_company(input_company)
        assert "value" in result
        # Should strip legal suffix
        assert "Corp" not in result["value"] or result["value"] == ""
        assert "Inc" not in result["value"] or result["value"] == ""

    @pytest.mark.parametrize("input_company", [
        "",
        "   ",
        None,
    ])
    def test_empty_input(self, input_company):
        """Test handling of empty inputs."""
        result = normalize_company(input_company)
        assert "value" in result

    def test_returns_dict_structure(self):
        """Test that result has expected structure."""
        result = normalize_company("Test Company Inc")
        assert isinstance(result, dict)
        assert "value" in result
        assert "confidence" in result


class TestCompanyToDomain:
    """Basic tests for company_to_domain."""

    @pytest.mark.parametrize("company,expected_type", [
        ("Google", dict),
        ("Microsoft Corporation", dict),
        ("Amazon.com Inc", dict),
    ])
    def test_returns_dict(self, company, expected_type):
        """Test that operator returns dict."""
        result = company_to_domain(company)
        assert isinstance(result, expected_type)
        assert "value" in result

    @pytest.mark.parametrize("company", [
        "",
        None,
        "   ",
    ])
    def test_empty_input(self, company):
        """Test handling of empty inputs."""
        result = company_to_domain(company)
        assert "value" in result


class TestEmailDomainMatch:
    """Basic tests for email_domain_match."""

    def test_matching_email_and_domain(self):
        """Test when email and domain match."""
        result = email_domain_match({
            "email": "john@acme.com",
            "company": "Acme Corp",
            "domain": "acme.com"
        })
        assert "value" in result
        assert isinstance(result["value"], bool)

    def test_string_input(self):
        """Test with string email input."""
        result = email_domain_match("john@acme.com")
        assert "value" in result
        assert isinstance(result["value"], bool)

    @pytest.mark.parametrize("input_data", [
        {},
        {"email": ""},
        {"email": None},
    ])
    def test_empty_input(self, input_data):
        """Test handling of empty inputs."""
        result = email_domain_match(input_data)
        assert "value" in result


class TestExtractDomainRoot:
    """Basic tests for extract_domain_root."""

    @pytest.mark.parametrize("input_url,has_domain", [
        ("https://www.example.com", True),
        ("example.com", True),
        ("subdomain.example.com", True),
        ("", False),
    ])
    def test_extracts_root(self, input_url, has_domain):
        """Test extraction of domain root."""
        result = extract_domain_root(input_url)
        assert "value" in result
        if has_domain:
            assert result["value"]
        else:
            assert result["value"] == ""


class TestExtractSubdomain:
    """Basic tests for extract_subdomain."""

    @pytest.mark.parametrize("input_url", [
        "https://www.example.com",
        "subdomain.example.com",
        "deep.subdomain.example.com",
        "example.com",
        "",
    ])
    def test_extracts_subdomain(self, input_url):
        """Test subdomain extraction."""
        result = extract_subdomain(input_url)
        assert "value" in result


class TestParseEmailPattern:
    """Basic tests for parse_email_pattern."""

    @pytest.mark.parametrize("email", [
        "first.last@company.com",
        "flast@company.com",
        "firstlast@company.com",
        "first_last@company.com",
        "",
    ])
    def test_parses_pattern(self, email):
        """Test email pattern parsing."""
        result = parse_email_pattern(email)
        assert "value" in result


class TestFixPhoneFormat:
    """Basic tests for fix_phone_format."""

    @pytest.mark.parametrize("phone", [
        "415-555-1234",
        "(415) 555-1234",
        "4155551234",
        "+1 415 555 1234",
        "",
    ])
    def test_formats_phone(self, phone):
        """Test phone formatting."""
        result = fix_phone_format(phone)
        assert "value" in result


class TestNormalizeJobFunction:
    """Basic tests for normalize_job_function."""

    @pytest.mark.parametrize("title", [
        "Software Engineer",
        "Sales Manager",
        "Marketing Director",
        "VP Engineering",
        "",
    ])
    def test_normalizes_function(self, title):
        """Test job function normalization."""
        result = normalize_job_function(title)
        assert "value" in result


class TestClassifyRevenueBand:
    """Basic tests for classify_revenue_band."""

    @pytest.mark.parametrize("revenue,expected_type", [
        (1000000, dict),
        (10000000, dict),
        (100000000, dict),
        (0, dict),
    ])
    def test_classifies_revenue(self, revenue, expected_type):
        """Test revenue band classification."""
        result = classify_revenue_band(revenue)
        assert isinstance(result, expected_type)
        assert "value" in result


class TestCalculateLeadScore:
    """Basic tests for calculate_lead_score."""

    def test_calculates_score(self):
        """Test lead score calculation."""
        data = {
            "email": "john@acme.com",
            "company_size": "100",
            "industry": "Software"
        }
        result = calculate_lead_score(data)
        assert "value" in result

    def test_empty_data(self):
        """Test with empty data."""
        result = calculate_lead_score({})
        assert "value" in result


class TestAssignTerritory:
    """Basic tests for assign_territory."""

    @pytest.mark.parametrize("data", [
        {"country": "US", "state": "CA"},
        {"country": "US", "state": "NY"},
        {"country": "UK"},
        {},
    ])
    def test_assigns_territory(self, data):
        """Test territory assignment."""
        result = assign_territory(data)
        assert "value" in result


class TestStateNameToCode:
    """Basic tests for state_name_to_code."""

    @pytest.mark.parametrize("state,expected_has_code", [
        ("California", True),
        ("New York", True),
        ("Texas", True),
        ("Invalid State", False),
        ("", False),
    ])
    def test_converts_state(self, state, expected_has_code):
        """Test state name to code conversion."""
        result = state_name_to_code(state)
        assert "value" in result


# Enrichment operators - basic smoke tests only
class TestEnrichmentOperators:
    """Smoke tests for enrichment operators that may call external APIs."""

    def test_enrich_company_country(self):
        """Test enrich_company_country doesn't crash."""
        from fmatch.saas.operators.enrich_company_country import enrich_company_country
        result = enrich_company_country({"company": "Test Company", "domain": "test.com"})
        assert "value" in result

    def test_enrich_company_headquarters(self):
        """Test enrich_company_headquarters doesn't crash."""
        from fmatch.saas.operators.enrich_company_headquarters import enrich_company_headquarters
        result = enrich_company_headquarters({"company": "Test Company", "domain": "test.com"})
        assert "value" in result

    def test_enrich_company_industry(self):
        """Test enrich_company_industry doesn't crash."""
        from fmatch.saas.operators.enrich_company_industry import enrich_company_industry
        result = enrich_company_industry({"company": "Test Company", "domain": "test.com"})
        assert "value" in result

    def test_enrich_social_handles(self):
        """Test enrich_social_handles doesn't crash."""
        from fmatch.saas.operators.enrich_social_handles import enrich_social_handles
        result = enrich_social_handles({"company": "Test Company", "domain": "test.com"})
        assert "value" in result

    def test_enrich_stock_info(self):
        """Test enrich_stock_info doesn't crash."""
        from fmatch.saas.operators.enrich_stock_info import enrich_stock_info
        result = enrich_stock_info({"company": "Test Company", "domain": "test.com"})
        assert "value" in result

    def test_enrich_subsidiaries(self):
        """Test enrich_subsidiaries doesn't crash."""
        from fmatch.saas.operators.enrich_subsidiaries import enrich_subsidiaries
        result = enrich_subsidiaries({"company": "Test Company", "domain": "test.com"})
        assert "value" in result


def run_tests():
    """Run all smoke tests and report results."""
    exit_code = pytest.main([__file__, "-v", "--tb=short"])

    if exit_code == 0:
        print("\n✅ All smoke tests passed!")
    else:
        print(f"\n❌ Smoke tests failed with exit code: {exit_code}")

    return exit_code


if __name__ == "__main__":
    exit(run_tests())



"""Tests for parse_address_components operator."""

import os
import sys

import pytest

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "..", "src"))

from fmatch.saas.operators.parse_address_components import parse_address_components


class TestParseAddressComponents:
    """Comprehensive tests for parse_address_components operator."""

    @pytest.mark.parametrize(
        "address,expected",
        [
            (
                "123 Main St, San Francisco, CA 94105",
                ["123 Main St", "San Francisco", "CA", "94105"],
            ),
            (
                "1600 Amphitheatre Pkwy, Mountain View, CA 94043",
                ["1600 Amphitheatre Pkwy", "Mountain View", "CA", "94043"],
            ),
            (
                "350 Fifth Avenue, New York, NY 10118",
                ["350 Fifth Avenue", "New York", "NY", "10118"],
            ),
        ],
    )
    def test_standard_us_addresses(self, address, expected):
        """Test parsing of standard US addresses."""
        result = parse_address_components(address)
        assert result["value"] == expected
        assert result["meta"]["confidence"] == pytest.approx(1.0)

    @pytest.mark.parametrize(
        "address,street,state",
        [
            ("10 Downing St, London SW1A 2AA, UK", "10 Downing St", "UK"),
            (
                "1 Avenue des Champs-Élysées, Paris 75008, France",
                "1 Avenue des Champs-Élysées",
                "France",
            ),
            (
                "Alexanderplatz 1, Berlin 10178, Germany",
                "Alexanderplatz 1",
                "Germany",
            ),
        ],
    )
    def test_international_addresses(self, address, street, state):
        """Test best-effort parsing of international addresses."""
        result = parse_address_components(address)
        values = result["value"]
        assert len(values) == 4
        assert values[0] == street
        assert values[2] == state
        # Best-effort parsing still returns strings
        assert all(isinstance(component, str) for component in values)

    @pytest.mark.parametrize(
        "address,expected",
        [
            ("", ["", "", "", ""]),
            ("   ", ["", "", "", ""]),
            ("Invalid", ["Invalid", "", "", ""]),
            ("Just a string", ["Just a string", "", "", ""]),
        ],
    )
    def test_edge_cases(self, address, expected):
        """Test handling of invalid/empty addresses."""
        result = parse_address_components(address)
        assert result["value"] == expected

    def test_none_input(self):
        """None input surfaces a null_input error via safe_transform."""
        result = parse_address_components(None)
        assert result["value"] is None
        assert result["error"] == "null_input"

    @pytest.mark.parametrize(
        "address",
        [
            "  123 Main St,  San Francisco,  CA  94105  ",
            "123 Main St,\tSan Francisco,\tCA\t94105",
            "123 Main St,\nSan Francisco,\nCA\n94105",
        ],
    )
    def test_whitespace_normalization(self, address):
        """Test that whitespace is normalized."""
        result = parse_address_components(address)
        assert result["value"] == ["123 Main St", "San Francisco", "CA", "94105"]

    def test_partial_addresses(self):
        """Test addresses with missing components."""
        result = parse_address_components("123 Main St, San Francisco, CA")
        assert result["value"] == ["123 Main St", "San Francisco", "CA", ""]

        result = parse_address_components("123 Main St, San Francisco")
        assert result["value"] == ["123 Main St", "San Francisco", "", ""]

    def test_two_part_city_state_postal_fallback(self):
        result = parse_address_components("123 Main St, Boston MA 02134")
        assert result["value"] == ["123 Main St", "Boston", "MA", "02134"]

    def test_city_postal_country_fallback(self):
        result = parse_address_components("123 Main St, London SW1A 2AA, UK")
        assert result["value"] == ["123 Main St", "London", "UK", "SW1A 2AA"]

    def test_apartment_numbers(self):
        """Test addresses with apartment/suite numbers."""
        result = parse_address_components("123 Main St Apt 4B, San Francisco, CA 94105")
        assert result["value"] == [
            "123 Main St Apt 4B",
            "San Francisco",
            "CA",
            "94105",
        ]

    def test_po_boxes(self):
        """Test P.O. Box addresses."""
        result = parse_address_components("PO Box 1234, San Francisco, CA 94105")
        assert result["value"] == ["PO Box 1234", "San Francisco", "CA", "94105"]

    @pytest.mark.parametrize(
        "address",
        [
            "123 O'Brien St, San José, CA 94105",
            "123 Main St #4, San Francisco, CA 94105",
            "123 Main St, Apt. 4-B, San Francisco, CA 94105",
        ],
    )
    def test_special_characters(self, address):
        """Test addresses with special characters."""
        result = parse_address_components(address)
        assert len(result["value"]) == 4
        assert all(isinstance(component, str) for component in result["value"])

    def test_output_structure(self):
        """Test that output structure is consistent."""
        result = parse_address_components("123 Main St, SF, CA 94105")
        assert isinstance(result, dict)
        assert "value" in result
        assert isinstance(result["value"], list)
        assert len(result["value"]) == 4
        for component in result["value"]:
            assert isinstance(component, str)

    def test_very_long_addresses(self):
        """Test handling of very long addresses."""
        long_address = "A" * 500 + ", " + "B" * 100 + ", CA 94105"
        result = parse_address_components(long_address)
        assert len(result["value"]) == 4
        assert result["value"][0].startswith("A" * 100)


def run_tests():
    """Run all tests and report results."""
    exit_code = pytest.main([__file__, "-v", "--tb=short"])

    if exit_code == 0:
        print("\nAll parse_address_components tests passed!")
    else:
        print(f"\nTests failed with exit code: {exit_code}")

    return exit_code


if __name__ == "__main__":
    raise SystemExit(run_tests())



"""
Enterprise dedupe planning helpers for large CRM jobs.

This module adds a blocking-first offline planning workflow that can operate on
multiple CRM exports (for example Contacts + Leads), emit review-ready merge
artifacts, and surface the engine's multi-algorithm matching profile in the
resulting job manifest.
"""

from __future__ import annotations

import asyncio
import csv
import json
import logging
import re
from collections import defaultdict
from dataclasses import dataclass, field
from difflib import SequenceMatcher
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple

LOGGER = logging.getLogger(__name__)

COMMON_CONSUMER_DOMAINS = {
    "gmail.com",
    "googlemail.com",
    "hotmail.com",
    "icloud.com",
    "outlook.com",
    "yahoo.com",
    "aol.com",
    "msn.com",
    "live.com",
    "proton.me",
    "protonmail.com",
}

CANONICAL_KEY_COLUMNS = {
    "name": "__match_name",
    "company": "__match_company",
    "email": "__match_email",
    "phone": "__match_phone",
    "domain": "__match_domain",
}

DEFAULT_KEY_COLUMNS = ["name", "email", "phone", "company", "domain"]
DEFAULT_BLOCKING_KEYS = ["email", "phone", "domain_name", "company_name"]
ALLOWED_BLOCKING_KEYS = set(DEFAULT_BLOCKING_KEYS)
DEFAULT_PROFILE = "default"
PROFILE_DEFAULTS = {
    "default": {
        "key_columns": DEFAULT_KEY_COLUMNS,
        "blocking_keys": DEFAULT_BLOCKING_KEYS,
        "threshold": 85,
        "notes": [
            "Balanced CRM dedupe profile using exact anchors first, with domain and company-name blocking enabled.",
            "Good for broad CRM cleanup, but domain-name blocking can be too permissive for B2B person resolution.",
        ],
    },
    "b2b_person": {
        "key_columns": DEFAULT_KEY_COLUMNS,
        "blocking_keys": ["email", "phone", "company_name"],
        "threshold": 90,
        "notes": [
            "B2B person dedupe profile favors exact person anchors first, with company-name as a secondary exact block.",
            "Domain stays in scoring but is removed from default blocking to avoid coworker chaining for same-name contacts.",
        ],
    },
}
ALLOWED_PROFILES = set(PROFILE_DEFAULTS)
FULL_SCAN_FALLBACK_MAX_ROWS = 250
NOISY_COMPANY_NAMES = {
    "self employed",
    "freelance",
    "freelancer",
    "n a",
    "na",
    "none",
    "unknown",
    "individual",
    "independent",
}

ENGINE_CAPABILITIES = {
    "multi_algorithm": True,
    "algorithm_families": [
        "Exact",
        "JaroWinkler",
        "Jaro",
        "TokenSetRatio",
        "TokenSortRatio",
        "WRatio",
        "PartialRatio",
        "Domain",
    ],
    "field_ensembles": {
        "email": ["Exact", "JaroWinkler", "WRatio"],
        "phone": ["Exact", "JaroWinkler"],
        "domain": ["TokenSetRatio", "WRatio", "JaroWinkler"],
        "name_company": ["WRatio", "TokenSetRatio", "PartialRatio"],
    },
    "notes": [
        "Blocking-first candidate generation keeps large CRM dedupe jobs tractable.",
        "Cross-object clusters are flagged for Lead-to-Contact resolution instead of blind merge.",
    ],
}


@dataclass(frozen=True)
class SourceSpec:
    object_type: str
    csv_path: Path


@dataclass
class Cluster:
    members: List[Dict[str, Any]]
    golden_index: int
    merge_confidence: float
    anchors: List[str] = field(default_factory=list)


class _UnionFind:
    def __init__(self, values: Iterable[Any]):
        self.parents = {value: value for value in values}

    def find(self, value: Any) -> Any:
        while self.parents[value] != value:
            self.parents[value] = self.parents[self.parents[value]]
            value = self.parents[value]
        return value

    def union(self, left: Any, right: Any) -> None:
        root_left = self.find(left)
        root_right = self.find(right)
        if root_left != root_right:
            self.parents[root_right] = root_left


def _normalize_text(value: Any) -> str:
    return re.sub(r"[^a-z0-9]+", " ", str(value or "").lower()).strip()


def _normalize_email(value: Any) -> str:
    return str(value or "").strip().lower()


def _normalize_phone(value: Any) -> str:
    digits = re.sub(r"\D+", "", str(value or ""))
    if len(digits) == 11 and digits.startswith("1"):
        return digits[1:]
    return digits


def _normalize_domain(value: Any) -> str:
    text = str(value or "").strip().lower()
    if not text:
        return ""
    if "@" in text and "/" not in text:
        text = text.split("@", 1)[1]
    text = re.sub(r"^https?://", "", text)
    text = text.split("/", 1)[0].split("?", 1)[0].split("#", 1)[0].strip()
    if text.startswith("www."):
        text = text[4:]
    return text


def _load_csv(csv_path: Path, *, limit: int | None = None) -> Tuple[List[Dict[str, str]], List[str]]:
    with csv_path.open(newline="", encoding="utf-8-sig") as handle:
        reader = csv.DictReader(handle)
        headers = list(reader.fieldnames or [])
        rows: List[Dict[str, str]] = []
        for row in reader:
            rows.append(row)
            if limit is not None and len(rows) >= limit:
                break
    if not rows:
        return [], headers
    return rows, headers


def _header_index(headers: Iterable[str]) -> Dict[str, str]:
    return {str(header).strip().lower(): str(header) for header in headers}


def _resolve_header(headers: Iterable[str], candidates: Sequence[str]) -> Optional[str]:
    header_map = _header_index(headers)
    for candidate in candidates:
        actual = header_map.get(candidate.lower())
        if actual:
            return actual
    return None


def _first_non_blank(row: Dict[str, Any], columns: Sequence[Optional[str]]) -> str:
    for column in columns:
        if not column:
            continue
        value = str(row.get(column, "") or "").strip()
        if value:
            return value
    return ""


def _compose_name(
    row: Dict[str, Any],
    *,
    name_column: Optional[str],
    first_name_column: Optional[str],
    last_name_column: Optional[str],
) -> str:
    direct = _first_non_blank(row, [name_column])
    if direct:
        return direct
    first_name = _first_non_blank(row, [first_name_column])
    last_name = _first_non_blank(row, [last_name_column])
    return " ".join(part for part in [first_name, last_name] if part).strip()


def _detect_columns(headers: Sequence[str]) -> Dict[str, Optional[str]]:
    return {
        "id": _resolve_header(headers, ["Id", "RecordId", "record_id", "ID", "sfid"]),
        "name": _resolve_header(headers, ["Name", "FullName", "Full Name", "Contact Name", "Lead Name"]),
        "first_name": _resolve_header(headers, ["FirstName", "First Name", "first_name", "given_name"]),
        "last_name": _resolve_header(headers, ["LastName", "Last Name", "last_name", "surname", "family_name"]),
        "company": _resolve_header(
            headers,
            ["Company", "Company Name", "Account Name", "AccountName", "Account", "Organization", "Org"],
        ),
        "email": _resolve_header(
            headers,
            ["Email", "Email Address", "EmailAddress", "work_email", "business_email", "email"],
        ),
        "phone": _resolve_header(
            headers,
            [
                "Phone",
                "Phone Number",
                "phone",
                "MobilePhone",
                "Mobile Phone",
                "Business Phone",
                "Work Phone",
            ],
        ),
        "domain": _resolve_header(
            headers,
            ["Domain", "Website", "Website Domain", "Company Domain", "web", "url", "URL"],
        ),
    }


def _prepare_source_rows(
    rows: Sequence[Dict[str, str]],
    headers: Sequence[str],
    source: SourceSpec,
) -> Tuple[List[Dict[str, Any]], Dict[str, Any]]:
    detected = _detect_columns(headers)
    prepared: List[Dict[str, Any]] = []
    for index, row in enumerate(rows, start=1):
        display_name = _compose_name(
            row,
            name_column=detected["name"],
            first_name_column=detected["first_name"],
            last_name_column=detected["last_name"],
        )
        display_company = _first_non_blank(row, [detected["company"]])
        display_email = _first_non_blank(row, [detected["email"]])
        display_phone = _first_non_blank(row, [detected["phone"]])
        raw_domain = _first_non_blank(row, [detected["domain"]])
        display_domain = _normalize_domain(raw_domain or display_email)
        record_id = _first_non_blank(row, [detected["id"]]) or f"{source.object_type}:{source.csv_path.stem}:{index}"

        prepared_row = dict(row)
        prepared_row.update(
            {
                "__object_type": source.object_type,
                "__source_file": source.csv_path.name,
                "__record_id": record_id,
                "__display_name": display_name,
                "__display_company": display_company,
                "__display_email": display_email,
                "__display_phone": display_phone,
                "__display_domain": display_domain,
                "__match_name": _normalize_text(display_name),
                "__match_company": _normalize_text(display_company),
                "__match_email": _normalize_email(display_email),
                "__match_phone": _normalize_phone(display_phone),
                "__match_domain": display_domain,
            }
        )
        prepared.append(prepared_row)

    source_meta = {
        "object_type": source.object_type,
        "path": str(source.csv_path),
        "rows": len(prepared),
        "detected_columns": detected,
    }
    return prepared, source_meta


def _resolve_key_columns(requested: Sequence[str], available_columns: Iterable[str]) -> List[str]:
    available_map = _header_index(available_columns)
    resolved: List[str] = []
    missing: List[str] = []
    for token in requested:
        key = str(token or "").strip()
        if not key:
            continue
        canonical = CANONICAL_KEY_COLUMNS.get(key.lower())
        if canonical:
            resolved.append(canonical)
            continue
        actual = available_map.get(key.lower())
        if actual:
            resolved.append(actual)
            continue
        missing.append(key)
    if missing:
        raise ValueError(f"Key columns not found: {', '.join(missing)}")
    if not resolved:
        raise ValueError("At least one match key column is required.")
    return resolved


def _resolve_profile_settings(
    *,
    profile: str,
    key_columns: Sequence[str] | None,
    blocking_keys: Sequence[str] | None,
    threshold: int | None,
) -> Dict[str, Any]:
    profile_name = str(profile or DEFAULT_PROFILE).strip().lower() or DEFAULT_PROFILE
    if profile_name not in ALLOWED_PROFILES:
        raise ValueError(f"Unsupported dedupe profile: {profile_name}")

    defaults = PROFILE_DEFAULTS[profile_name]
    key_column_source = defaults["key_columns"] if key_columns is None else key_columns
    blocking_key_source = defaults["blocking_keys"] if blocking_keys is None else blocking_keys
    resolved_key_columns = [str(key).strip() for key in key_column_source if str(key).strip()]
    resolved_blocking_keys = [
        str(key).strip().lower() for key in blocking_key_source if str(key).strip()
    ]
    resolved_threshold = int(defaults["threshold"] if threshold is None else threshold)
    return {
        "profile": profile_name,
        "key_columns": resolved_key_columns,
        "blocking_keys": resolved_blocking_keys,
        "threshold": resolved_threshold,
        "notes": list(defaults.get("notes") or []),
        "defaults_applied": {
            "key_columns": key_columns is None,
            "blocking_keys": blocking_keys is None,
            "threshold": threshold is None,
        },
    }


def _candidate_anchor_keys(row: Dict[str, Any], blocking_keys: Sequence[str]) -> List[Tuple[str, str]]:
    anchors: List[Tuple[str, str]] = []
    name = row.get("__match_name", "") or ""
    company = row.get("__match_company", "") or ""
    email = row.get("__match_email", "") or ""
    phone = row.get("__match_phone", "") or ""
    domain = row.get("__match_domain", "") or ""

    if "email" in blocking_keys and email:
        anchors.append(("email", email))
    if "phone" in blocking_keys and len(phone) >= 7:
        anchors.append(("phone", phone))
    if "domain_name" in blocking_keys and domain and name and domain not in COMMON_CONSUMER_DOMAINS:
        anchors.append(("domain_name", f"{domain}|{name}"))
    if "company_name" in blocking_keys and company and name and company not in NOISY_COMPANY_NAMES:
        anchors.append(("company_name", f"{company}|{name}"))
    return anchors


def _build_candidate_components(
    rows: Sequence[Dict[str, Any]],
    blocking_keys: Sequence[str],
    block_size_cap: int,
) -> Tuple[List[List[int]], Dict[str, Any], Dict[int, set[str]]]:
    anchor_groups: Dict[Tuple[str, str], List[int]] = defaultdict(list)
    row_anchor_kinds: Dict[int, set[str]] = defaultdict(set)
    active_rows_by_kind: Dict[str, set[int]] = defaultdict(set)
    group_counts: Dict[str, int] = defaultdict(int)
    ignored_counts: Dict[str, int] = defaultdict(int)
    ignored_examples: List[Dict[str, Any]] = []

    for row in rows:
        row_index = int(row.get("__row_index", -1))
        if row_index < 0:
            continue
        seen_for_row: set[Tuple[str, str]] = set()
        for anchor in _candidate_anchor_keys(row, blocking_keys):
            if anchor in seen_for_row:
                continue
            seen_for_row.add(anchor)
            anchor_groups[anchor].append(row_index)

    union_find = _UnionFind(range(len(rows)))

    for (anchor_kind, anchor_value), member_ids in anchor_groups.items():
        unique_ids = sorted(set(member_ids))
        if len(unique_ids) < 2:
            continue
        if len(unique_ids) > block_size_cap:
            ignored_counts[anchor_kind] += 1
            if len(ignored_examples) < 10:
                ignored_examples.append(
                    {"anchor": anchor_kind, "size": len(unique_ids), "sample": str(anchor_value)[:80]}
                )
            continue

        group_counts[anchor_kind] += 1
        active_rows_by_kind[anchor_kind].update(unique_ids)
        for member_id in unique_ids:
            row_anchor_kinds[member_id].add(anchor_kind)
        for member_id in unique_ids[1:]:
            union_find.union(unique_ids[0], member_id)

    components_map: Dict[int, List[int]] = defaultdict(list)
    for member_id in sorted(row_anchor_kinds):
        components_map[union_find.find(member_id)].append(member_id)

    components = [sorted(member_ids) for member_ids in components_map.values() if len(member_ids) > 1]
    fallback_full_scan = False
    if not components and len(rows) <= FULL_SCAN_FALLBACK_MAX_ROWS:
        components = [list(range(len(rows)))] if len(rows) > 1 else []
        fallback_full_scan = bool(components)
        for idx in range(len(rows)):
            row_anchor_kinds.setdefault(idx, set())

    summary = {
        "blocking_keys": list(blocking_keys),
        "candidate_components": len(components),
        "candidate_rows": len({member_id for component in components for member_id in component}),
        "largest_component": max((len(component) for component in components), default=0),
        "fallback_full_scan": fallback_full_scan,
        "anchors": {
            key: {
                "groups": group_counts.get(key, 0),
                "rows": len(active_rows_by_kind.get(key, set())),
                "ignored_groups": ignored_counts.get(key, 0),
            }
            for key in blocking_keys
        },
        "ignored_examples": ignored_examples,
    }
    return components, summary, row_anchor_kinds


def _sequence_similarity(left: str, right: str) -> float:
    return SequenceMatcher(None, left, right).ratio() * 100.0


def _column_similarity(column: str, left: Any, right: Any) -> Optional[float]:
    left_text = str(left or "").strip()
    right_text = str(right or "").strip()
    if not left_text or not right_text:
        return None

    lowered = column.lower()
    if "email" in lowered:
        left_email = _normalize_email(left_text)
        right_email = _normalize_email(right_text)
        return 100.0 if left_email == right_email else _sequence_similarity(left_email, right_email)

    if "phone" in lowered:
        left_phone = _normalize_phone(left_text)
        right_phone = _normalize_phone(right_text)
        if not left_phone or not right_phone:
            return None
        if left_phone == right_phone:
            return 100.0
        if len(left_phone) >= 7 and len(right_phone) >= 7 and left_phone[-7:] == right_phone[-7:]:
            return 92.0
        return _sequence_similarity(left_phone, right_phone)

    if any(token in lowered for token in ("domain", "website", "url", "web")):
        left_domain = _normalize_domain(left_text)
        right_domain = _normalize_domain(right_text)
        if not left_domain or not right_domain:
            return None
        if left_domain == right_domain:
            return 100.0

        left_parts = left_domain.split(".")
        right_parts = right_domain.split(".")
        left_registrable = ".".join(left_parts[-2:]) if len(left_parts) >= 2 else left_domain
        right_registrable = ".".join(right_parts[-2:]) if len(right_parts) >= 2 else right_domain
        if left_registrable == right_registrable:
            return 95.0
        return _sequence_similarity(left_domain, right_domain)

    left_norm = _normalize_text(left_text)
    right_norm = _normalize_text(right_text)
    if not left_norm or not right_norm:
        return None
    return _sequence_similarity(left_norm, right_norm)


def _row_similarity(left: Dict[str, Any], right: Dict[str, Any], key_columns: Sequence[str]) -> float:
    scores: List[float] = []
    has_exact_anchor = False
    has_domain_anchor = False

    for column in key_columns:
        score = _column_similarity(column, left.get(column), right.get(column))
        if score is None:
            continue
        scores.append(score)
        lowered = column.lower()
        if "email" in lowered and score >= 100.0:
            has_exact_anchor = True
        elif "phone" in lowered and score >= 100.0:
            has_exact_anchor = True
        elif any(token in lowered for token in ("domain", "website", "url", "web")) and score >= 95.0:
            has_domain_anchor = True

    if not scores:
        return 0.0

    blended = sum(scores) / len(scores)
    if has_exact_anchor:
        return max(blended, 98.0)
    if has_domain_anchor and left.get("__match_name") and left.get("__match_name") == right.get("__match_name"):
        return max(blended, 90.0)
    return blended


def _choose_golden(members: Sequence[Dict[str, Any]], survivor_policy: str) -> int:
    best_index = 0
    best_score = float("-inf")
    for index, row in enumerate(members):
        score = 0.0
        for key, value in row.items():
            if key.startswith("__"):
                continue
            if str(value or "").strip():
                score += 1.0
        if row.get("__display_email"):
            score += 4.0
        if row.get("__display_phone"):
            score += 2.5
        if row.get("__display_domain"):
            score += 2.0
        if row.get("__display_company"):
            score += 1.0
        if survivor_policy == "crm_person":
            object_type = str(row.get("__object_type", "")).strip().lower()
            if object_type == "contact":
                score += 8.0
            elif object_type == "lead":
                score += 3.0
        score += 0.001 * index
        if score > best_score:
            best_score = score
            best_index = index
    return best_index


def _cluster_rows(
    rows: Sequence[Dict[str, Any]],
    key_columns: Sequence[str],
    threshold: int,
    survivor_policy: str,
) -> List[Cluster]:
    union_find = _UnionFind(range(len(rows)))

    for left_index in range(len(rows)):
        for right_index in range(left_index + 1, len(rows)):
            similarity = _row_similarity(rows[left_index], rows[right_index], key_columns)
            if similarity >= threshold:
                union_find.union(left_index, right_index)

    groups: Dict[int, List[Dict[str, Any]]] = defaultdict(list)
    for index, row in enumerate(rows):
        groups[union_find.find(index)].append(row)

    clusters: List[Cluster] = []
    for members in groups.values():
        golden_index = _choose_golden(members, survivor_policy=survivor_policy)
        clusters.append(Cluster(members=list(members), golden_index=golden_index, merge_confidence=1.0))
    return clusters


def _extract_pairs(matches_df: Any, id_field: str) -> List[Tuple[str, str]]:
    try:
        import pandas as pd  # type: ignore
    except Exception as exc:  # pragma: no cover - optional dependency
        raise RuntimeError("pandas is required for engine dedup.") from exc

    pairs: List[Tuple[str, str]] = []
    columns = list(getattr(matches_df, "columns", []))
    id_like = [column for column in columns if "id" in str(column).lower()]

    def _append_pairs(left_column: str, right_column: str) -> None:
        for _, row in matches_df.iterrows():
            left_value = row.get(left_column)
            right_value = row.get(right_column)
            if pd.isna(left_value) or pd.isna(right_value):
                continue
            left_text = str(left_value).strip()
            right_text = str(right_value).strip()
            if not left_text or not right_text or left_text == right_text:
                continue
            pairs.append((left_text, right_text))

    if id_field in columns and len(id_like) >= 2:
        for column in id_like:
            if column == id_field:
                continue
            _append_pairs(id_field, column)
        if pairs:
            return pairs

    if len(id_like) >= 2:
        _append_pairs(id_like[0], id_like[1])

    return pairs


def _clusters_from_pairs(
    rows: Sequence[Dict[str, Any]],
    pairs: Sequence[Tuple[str, str]],
    survivor_policy: str,
) -> List[Cluster]:
    id_to_row = {str(row["__row_index"]): row for row in rows}
    union_find = _UnionFind(id_to_row)

    for left, right in pairs:
        if left in id_to_row and right in id_to_row:
            union_find.union(left, right)

    groups: Dict[str, List[Dict[str, Any]]] = defaultdict(list)
    for row_id, row in id_to_row.items():
        groups[union_find.find(row_id)].append(row)

    clusters: List[Cluster] = []
    for members in groups.values():
        golden_index = _choose_golden(members, survivor_policy=survivor_policy)
        clusters.append(Cluster(members=list(members), golden_index=golden_index, merge_confidence=1.0))
    return clusters


def _build_engine_mappings(key_columns: Sequence[str]) -> List[Dict[str, Any]]:
    mappings: List[Dict[str, Any]] = []
    for column in key_columns:
        field_name = column.lower()
        mapping: Dict[str, Any] = {"source": column, "reference": column, "weight": 1.0}
        if "email" in field_name:
            mapping["algorithms"] = [
                {"name": "Exact", "weight": 0.5},
                {"name": "JaroWinkler", "weight": 0.3},
                {"name": "WRatio", "weight": 0.2},
            ]
        elif "phone" in field_name:
            mapping["algorithms"] = [
                {"name": "Exact", "weight": 0.6},
                {"name": "JaroWinkler", "weight": 0.4},
            ]
        elif any(token in field_name for token in ("domain", "website", "url", "web")):
            mapping["algorithms"] = [
                {"name": "TokenSetRatio", "weight": 0.5},
                {"name": "WRatio", "weight": 0.3},
                {"name": "JaroWinkler", "weight": 0.2},
            ]
        elif any(token in field_name for token in ("name", "company", "account", "org")):
            mapping["algorithms"] = [
                {"name": "WRatio", "weight": 0.5},
                {"name": "TokenSetRatio", "weight": 0.3},
                {"name": "PartialRatio", "weight": 0.2},
            ]
        else:
            mapping["algorithms"] = [
                {"name": "WRatio", "weight": 0.6},
                {"name": "TokenSetRatio", "weight": 0.4},
            ]
        mapping["preferred_algo"] = mapping["algorithms"][0]["name"]
        mappings.append(mapping)
    return mappings


def _engine_cluster(
    rows: Sequence[Dict[str, Any]],
    key_columns: Sequence[str],
    threshold: int,
    survivor_policy: str,
) -> List[Cluster]:
    try:
        import pandas as pd  # type: ignore
        from fmatch.saas.services.match_gateway import MatchGateway  # type: ignore
    except Exception as exc:  # pragma: no cover - optional dependency
        raise RuntimeError("MatchGateway is not available") from exc

    df = pd.DataFrame(rows)
    id_field = "__source_id"
    df[id_field] = [str(row["__row_index"]) for row in rows]
    mappings = _build_engine_mappings(key_columns)

    async def _run() -> Any:
        gateway = await MatchGateway.create()
        return await gateway.dedup(df, threshold=threshold, mappings=mappings)

    try:
        result = asyncio.run(_run())
    except Exception as exc:  # pragma: no cover - runtime depends on engine env
        raise RuntimeError(f"Gateway dedup failed: {exc}") from exc

    matches_df = getattr(result, "matches", None)
    if matches_df is None:
        raise RuntimeError("Engine dedup did not return matches DataFrame.")

    pairs = _extract_pairs(matches_df, id_field)
    return _clusters_from_pairs(rows, pairs, survivor_policy=survivor_policy)


def _recommended_action(members: Sequence[Dict[str, Any]]) -> str:
    object_types = {str(member.get("__object_type", "")).strip().lower() for member in members}
    if "lead" in object_types and "contact" in object_types:
        return "lead_to_contact_resolution"
    return "merge"


def _cluster_type(members: Sequence[Dict[str, Any]]) -> str:
    object_types = {str(member.get("__object_type", "")).strip().lower() for member in members}
    if len(object_types) <= 1:
        return "same_object"
    if "lead" in object_types and "contact" in object_types:
        return "lead_to_contact"
    return "cross_object"


def _estimate_confidence(anchors: Sequence[str], size: int) -> float:
    confidence = 0.8
    if "email" in anchors:
        confidence = 0.99
    elif "phone" in anchors:
        confidence = 0.95
    elif "domain_name" in anchors:
        confidence = 0.88
    elif "company_name" in anchors:
        confidence = 0.84

    if size >= 3 and confidence < 0.99:
        confidence = min(confidence + 0.02, 0.99)
    return round(confidence, 2)


def _has_low_name_similarity(anchor_record: Dict[str, Any], members: Sequence[Dict[str, Any]]) -> bool:
    anchor_name = str(anchor_record.get("__match_name", "") or "").strip()
    if not anchor_name:
        return False
    for member in members:
        if member is anchor_record:
            continue
        member_name = str(member.get("__match_name", "") or "").strip()
        if not member_name:
            continue
        if _sequence_similarity(anchor_name, member_name) < 55.0:
            return True
    return False


def _review_notes(
    members: Sequence[Dict[str, Any]],
    anchors: Sequence[str],
    *,
    golden_record: Dict[str, Any],
) -> List[str]:
    notes: List[str] = []
    action = _recommended_action(members)
    if action == "lead_to_contact_resolution":
        notes.append(
            "Mixed Lead + Contact cluster; resolve with Lead-to-Contact review instead of blind merge."
        )
    if "email" in anchors and _has_low_name_similarity(golden_record, members):
        notes.append("Exact email anchor but divergent names detected; verify this is not a shared inbox.")
    if "email" not in anchors and "phone" not in anchors:
        notes.append("No exact email/phone anchor; keep survivor approval behind a human review gate.")
    if len(members) >= 5:
        notes.append("Large duplicate cluster; validate downstream ownership before execution.")
    return notes


def _public_record(row: Dict[str, Any]) -> Dict[str, Any]:
    payload = {key: value for key, value in row.items() if not key.startswith("__")}
    payload.update(
        {
            "record_id": row.get("__record_id", ""),
            "object_type": row.get("__object_type", ""),
            "source_file": row.get("__source_file", ""),
            "match_name": row.get("__display_name", ""),
            "match_company": row.get("__display_company", ""),
            "match_email": row.get("__display_email", ""),
            "match_phone": row.get("__display_phone", ""),
            "match_domain": row.get("__display_domain", ""),
        }
    )
    return payload


def _cluster_payload(
    cluster: Cluster,
    row_anchor_kinds: Dict[int, set[str]],
    survivor_policy: str,
    cluster_id: str,
) -> Dict[str, Any]:
    member_indexes = [int(member.get("__row_index", -1)) for member in cluster.members]
    anchors = sorted({kind for idx in member_indexes for kind in row_anchor_kinds.get(idx, set())})
    public_members = [_public_record(member) for member in cluster.members]
    golden_record = cluster.members[cluster.golden_index]
    return {
        "cluster_id": cluster_id,
        "cluster_type": _cluster_type(cluster.members),
        "recommended_action": _recommended_action(cluster.members),
        "survivor_policy": survivor_policy,
        "object_types": sorted({member.get("object_type", "") for member in public_members}),
        "anchors": anchors,
        "merge_confidence": _estimate_confidence(anchors, len(public_members)),
        "review_notes": _review_notes(cluster.members, anchors, golden_record=golden_record),
        "members": public_members,
        "golden_record": public_members[cluster.golden_index],
    }


def _write_review_worksheet(plan: Dict[str, Any], output_path: Path) -> Path:
    output_path.parent.mkdir(parents=True, exist_ok=True)
    fieldnames = [
        "cluster_id",
        "cluster_type",
        "recommended_action",
        "merge_confidence",
        "record_id",
        "object_type",
        "is_survivor",
        "source_file",
        "anchors",
        "match_name",
        "match_company",
        "match_email",
        "match_phone",
        "match_domain",
        "review_notes",
    ]
    with output_path.open("w", newline="", encoding="utf-8") as handle:
        writer = csv.DictWriter(handle, fieldnames=fieldnames)
        writer.writeheader()
        for cluster in plan.get("clusters", []):
            survivor_id = str(cluster.get("golden_record", {}).get("record_id", ""))
            anchors = ",".join(cluster.get("anchors", []))
            notes = " | ".join(cluster.get("review_notes", []))
            for member in cluster.get("members", []):
                writer.writerow(
                    {
                        "cluster_id": cluster.get("cluster_id", ""),
                        "cluster_type": cluster.get("cluster_type", ""),
                        "recommended_action": cluster.get("recommended_action", ""),
                        "merge_confidence": cluster.get("merge_confidence", ""),
                        "record_id": member.get("record_id", ""),
                        "object_type": member.get("object_type", ""),
                        "is_survivor": str(member.get("record_id", "")) == survivor_id,
                        "source_file": member.get("source_file", ""),
                        "anchors": anchors,
                        "match_name": member.get("match_name", ""),
                        "match_company": member.get("match_company", ""),
                        "match_email": member.get("match_email", ""),
                        "match_phone": member.get("match_phone", ""),
                        "match_domain": member.get("match_domain", ""),
                        "review_notes": notes,
                    }
                )
    return output_path


def build_enterprise_dedupe_plan(
    source_specs: Sequence[Tuple[str, Path] | SourceSpec],
    *,
    key_columns: Optional[Sequence[str]] = None,
    threshold: int | None = None,
    output_path: Path,
    review_output_path: Optional[Path] = None,
    overwrite: bool = False,
    local: bool = False,
    survivor_policy: str = "crm_person",
    blocking_keys: Optional[Sequence[str]] = None,
    profile: str = DEFAULT_PROFILE,
    block_size_cap: int = 5000,
    limit_per_source: int | None = None,
    pairwise_warning_threshold: int = 1500,
) -> Dict[str, Any]:
    if output_path.exists() and not overwrite:
        raise FileExistsError(f"Output {output_path} exists. Use --overwrite.")
    if review_output_path and review_output_path.exists() and not overwrite:
        raise FileExistsError(f"Output {review_output_path} exists. Use --overwrite.")

    normalized_sources = [
        source if isinstance(source, SourceSpec) else SourceSpec(object_type=source[0], csv_path=source[1])
        for source in source_specs
    ]
    if not normalized_sources:
        raise ValueError("At least one source export is required.")

    resolved_profile = _resolve_profile_settings(
        profile=profile,
        key_columns=key_columns,
        blocking_keys=blocking_keys,
        threshold=threshold,
    )
    requested_key_columns = list(resolved_profile["key_columns"])
    normalized_blocking_keys = list(resolved_profile["blocking_keys"])
    resolved_threshold = int(resolved_profile["threshold"])

    invalid_blocking_keys = sorted(set(normalized_blocking_keys) - ALLOWED_BLOCKING_KEYS)
    if invalid_blocking_keys:
        raise ValueError(f"Unsupported blocking keys: {', '.join(invalid_blocking_keys)}")
    if not normalized_blocking_keys:
        raise ValueError("At least one blocking key is required.")
    if block_size_cap < 2:
        raise ValueError("--block-size-cap must be at least 2.")
    if limit_per_source is not None and limit_per_source < 1:
        raise ValueError("--limit must be at least 1 when provided.")
    if pairwise_warning_threshold < 2:
        raise ValueError("pairwise_warning_threshold must be at least 2.")

    all_rows: List[Dict[str, Any]] = []
    sources_meta: List[Dict[str, Any]] = []
    warnings: List[str] = []

    for source in normalized_sources:
        rows, headers = _load_csv(source.csv_path, limit=limit_per_source)
        if not rows:
            raise ValueError(f"Input CSV is empty: {source.csv_path}")
        prepared_rows, source_meta = _prepare_source_rows(rows, headers, source)
        if limit_per_source is not None:
            source_meta["row_limit_applied"] = limit_per_source
        sources_meta.append(source_meta)
        all_rows.extend(prepared_rows)

    for row_index, row in enumerate(all_rows):
        row["__row_index"] = row_index

    resolved_key_columns = _resolve_key_columns(requested_key_columns, all_rows[0].keys())
    components, blocking_summary, row_anchor_kinds = _build_candidate_components(
        all_rows, normalized_blocking_keys, block_size_cap
    )

    used_modes: set[str] = set()
    duplicate_clusters: List[Cluster] = []
    engine_requested = not local

    for component in components:
        component_rows = [all_rows[index] for index in component]
        component_clusters: Optional[List[Cluster]] = None
        if engine_requested:
            try:
                component_clusters = _engine_cluster(
                    component_rows,
                    key_columns=resolved_key_columns,
                    threshold=resolved_threshold,
                    survivor_policy=survivor_policy,
                )
                used_modes.add("engine")
            except Exception as exc:
                warnings.append(f"Engine fallback on candidate block ({len(component_rows)} rows): {exc}")

        if component_clusters is None:
            if len(component_rows) > pairwise_warning_threshold:
                warnings.append(
                    "Large local candidate block encountered "
                    f"({len(component_rows)} rows). Tighten blocking keys or use engine mode for this segment."
                )
            component_clusters = _cluster_rows(
                component_rows,
                key_columns=resolved_key_columns,
                threshold=resolved_threshold,
                survivor_policy=survivor_policy,
            )
            used_modes.add("local")

        duplicate_clusters.extend(cluster for cluster in component_clusters if len(cluster.members) > 1)

    payload_clusters = [
        _cluster_payload(
            cluster,
            row_anchor_kinds=row_anchor_kinds,
            survivor_policy=survivor_policy,
            cluster_id=f"crm_{index:04d}",
        )
        for index, cluster in enumerate(duplicate_clusters, start=1)
    ]
    payload_clusters.sort(
        key=lambda cluster: (
            -float(cluster.get("merge_confidence", 0.0)),
            -len(cluster.get("members", [])),
            str(cluster.get("cluster_id", "")),
        )
    )

    if blocking_summary.get("ignored_examples"):
        warnings.append(
            "Some blocking groups were skipped because they exceeded the block-size cap. "
            "Review stats.blocking.ignored_examples before execution."
        )

    object_counts: Dict[str, int] = defaultdict(int)
    for row in all_rows:
        object_counts[str(row.get("__object_type", "") or "unknown")] += 1

    action_counts: Dict[str, int] = defaultdict(int)
    for cluster in payload_clusters:
        action_counts[str(cluster.get("recommended_action", "merge"))] += 1

    if not used_modes:
        mode_used = "local" if local else "engine"
    elif used_modes == {"engine"}:
        mode_used = "engine"
    elif used_modes == {"local"}:
        mode_used = "local"
    else:
        mode_used = "mixed"

    plan = {
        "$schema": "MergePlan/v2",
        "job_type": "crm_dedupe",
        "sources": sources_meta,
        "matching_profile": {
            "profile": resolved_profile["profile"],
            "profile_defaults_applied": dict(resolved_profile["defaults_applied"]),
            "profile_notes": list(resolved_profile["notes"]),
            "mode_requested": "local" if local else "engine",
            "mode_used": mode_used,
            "threshold": resolved_threshold,
            "candidate_strategy": "blocking_first",
            "key_columns": requested_key_columns,
            "resolved_key_columns": resolved_key_columns,
            "blocking_keys": normalized_blocking_keys,
            "limit_per_source": limit_per_source,
            "capabilities": ENGINE_CAPABILITIES,
        },
        "stats": {
            "records_total": len(all_rows),
            "candidate_rows": blocking_summary.get("candidate_rows", 0),
            "candidate_components": blocking_summary.get("candidate_components", 0),
            "largest_component": blocking_summary.get("largest_component", 0),
            "clusters_found": len(payload_clusters),
            "duplicates_identified": sum(max(len(cluster.get("members", [])) - 1, 0) for cluster in payload_clusters),
            "object_counts": dict(sorted(object_counts.items())),
            "actions": dict(sorted(action_counts.items())),
            "blocking": blocking_summary,
        },
        "warnings": warnings,
        "clusters": payload_clusters,
    }

    output_path.parent.mkdir(parents=True, exist_ok=True)
    output_path.write_text(json.dumps(plan, indent=2), encoding="utf-8")

    if review_output_path:
        _write_review_worksheet(plan, review_output_path)

    return plan



"""Centralized B2C email/root domains for server-side checks."""

B2C_TOP = {
    "gmail.com",
    "yahoo.com",
    "hotmail.com",
    "outlook.com",
    "live.com",
    "aol.com",
    "icloud.com",
    "me.com",
    "protonmail.com",
    "proton.me",
    "zoho.com",
    "gmx.com",
    "mail.com",
    "mail.ru",
    "qq.com",
    "163.com",
    "fastmail.com",
}



"""
Quota enforcement middleware with grace band support.
Protects both your infra and user's SFDC org limits.
"""

from typing import Dict, Any, Optional
from datetime import datetime, timezone
from fastapi import HTTPException
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy import select, update
import logging

log = logging.getLogger(__name__)


# =============================================================================
# Tier limits configuration - NOW LOADED FROM CENTRALIZED CONFIG
# =============================================================================
# This imports from config/tier_limits.yaml via the config loader.
# If you need to change tier allocations, update the YAML file instead of
# hardcoding values here. This ensures backend, website, and docs stay in sync.

try:
    from ..config.tier_limits import get_tier_limits

    TIER_LIMITS = get_tier_limits()
    log.info(
        f"Loaded tier limits from centralized config: {list(TIER_LIMITS.keys())}"
    )
except Exception as e:
    # Fallback to minimal config if YAML is missing (prevents startup failure)
    log.error(
        f"Failed to load tier limits from config/tier_limits.yaml: {e}\n"
        f"Using fallback minimal configuration. This should only happen in dev!"
    )
    # Fallback values synced with config/tier_limits.yaml v1.2.0 (2026-01-19)
    # Changes: users=1 for all tiers, max_schedules=5/25/100, enrichments shelved
    TIER_LIMITS = {
        "free": {
            "monthly_rows": 5000,
            "rows_per_run": 5000,
            "grace_percentage": 0.10,
            "enrichment_credits": 0,  # Shelved for V1
            "max_schedules": 5,
            "allowed_frequencies": ["daily", "weekly"],
            "concurrent_jobs": 1,
            "candidate_cap": 50,
            "require_b2c": True,
            "allow_advanced_algos": False,
            "soql_runs_per_month": 15,
            "salesforce_orgs": 1,
            "users": 1,
            # AI Limits
            "gremlin_enabled": False,
            "gremlin_messages_per_day": 0,
            "ai_report_row_cap": 5000,
        },
        "pro": {
            "monthly_rows": 100000,
            "rows_per_run": 50000,
            "grace_percentage": 0.05,
            "enrichment_credits": 0,  # Shelved for V1
            "max_schedules": 25,
            "allowed_frequencies": ["hourly", "daily", "weekly"],
            "concurrent_jobs": 1,
            "candidate_cap": 2000,
            "require_b2c": False,
            "allow_advanced_algos": True,
            "soql_runs_per_month": 100,
            "salesforce_orgs": 1,
            "users": 1,  # Seats sold as add-ons at $34/mo
            # AI Limits
            "gremlin_enabled": False,              # Gremlin Chat shelved
            "gremlin_messages_per_day": 0,
            "ai_report_row_cap": 50000,
        },
        "scale": {
            "monthly_rows": 500000,
            "rows_per_run": 100000,
            "grace_percentage": 0,
            "enrichment_credits": 0,  # Shelved for V1
            "max_schedules": 100,
            "allowed_frequencies": ["15min", "hourly", "daily", "weekly", "cron"],
            "concurrent_jobs": 4,
            "candidate_cap": 5000,
            "require_b2c": False,
            "allow_advanced_algos": True,
            "soql_runs_per_month": None,
            "salesforce_orgs": 1,  # Multi-org postponed to enterprise
            "users": 1,  # Seats sold as add-ons at $34/mo
            # AI Limits
            "gremlin_enabled": False,              # Gremlin Chat shelved
            "gremlin_messages_per_day": 0,
            "ai_report_row_cap": 100000,
        },
        "unleashed": {
            "monthly_rows": 2000000,
            "rows_per_run": 500000,
            "grace_percentage": 0,
            "enrichment_credits": 0,  # Shelved for V1
            "max_schedules": 500,
            "allowed_frequencies": ["15min", "hourly", "daily", "weekly", "cron"],
            "concurrent_jobs": 8,
            "candidate_cap": 50000,
            "require_b2c": False,
            "allow_advanced_algos": True,
            "soql_runs_per_month": None,
            "salesforce_orgs": 5,
            "users": 1,  # Seats sold as add-ons at $34/mo
            # AI Limits
            "gremlin_enabled": False,              # Gremlin Chat shelved
            "gremlin_messages_per_day": 0,
            "ai_report_row_cap": 500000,
            # BYO Enrichment - BUNDLED with unlimited limits
            "byo_enrichment_enabled": True,
            "byo_bundled": True,
            "byo_providers": ["foundrygraph", "fullenrich", "pdl"],
            "byo_max_rows_per_job": 1000,
            "byo_concurrent_enrichment_jobs": 10,
            "byo_daily_enrichment_rows": 50000,
            "byo_monthly_enrichment_rows": None,  # Unlimited
            "byo_scheduled_enrichment": True,
            "byo_priority_queue": True,
        },
    }
    TIER_LIMITS["business"] = TIER_LIMITS["scale"].copy()

    # BYO Provider Locker add-on limits (for Pro/Scale with add-on purchased)
    # Synced with config/enrichment_pricing.yaml addons.byo_provider_locker
    BYO_ADDON_LIMITS = {
        "pro": {
            "max_rows_per_job": 250,
            "concurrent_enrichment_jobs": 1,
            "daily_enrichment_rows": 1000,
            "monthly_enrichment_rows": 5000,
            "scheduled_enrichment": False,
            "providers": ["foundrygraph", "fullenrich", "pdl"],
        },
        "scale": {
            "max_rows_per_job": 500,
            "concurrent_enrichment_jobs": 2,
            "daily_enrichment_rows": 5000,
            "monthly_enrichment_rows": 25000,
            "scheduled_enrichment": True,
            "scheduled_frequencies": ["hourly", "daily"],
            "providers": ["foundrygraph", "fullenrich", "pdl"],
        },
    }


class QuotaExceeded(HTTPException):
    """Quota limit exceeded with upgrade prompt."""

    def __init__(
        self, message: str, limit_type: str, current: int, limit: int, tier: str
    ):
        super().__init__(
            status_code=429,
            detail={
                "error": "quota_exceeded",
                "message": message,
                "limit_type": limit_type,
                "current": current,
                "limit": limit,
                "tier": tier,
                "upgrade_url": "/pricing",
                "upgrade_message": f"Upgrade to Pro for {TIER_LIMITS['pro'][limit_type]} {limit_type}/day",
            },
        )


class UsageQuota:
    """Usage quota tracking with grace band."""

    def __init__(
        self,
        account_id: str,
        tier: str = "free",
        monthly_rows_used: int = 0,
        enrichment_credits_used: int = 0,
        active_schedules: int = 0,
        grace_uses_remaining: int = 1,
        last_grace_at: Optional[datetime] = None,
        reset_at: Optional[datetime] = None,
        timezone: str = "UTC",
        # Legacy fields for backwards compatibility
        daily_runs_used: int = 0,
        daily_rows_used: int = 0,
    ):
        self.account_id = account_id
        self.tier = tier
        self.monthly_rows_used = monthly_rows_used
        self.enrichment_credits_used = enrichment_credits_used
        self.active_schedules = active_schedules
        self.grace_uses_remaining = grace_uses_remaining
        self.last_grace_at = last_grace_at
        self.reset_at = reset_at or self._next_reset(timezone)
        self.timezone = timezone
        # Legacy compatibility
        self.daily_runs_used = daily_runs_used
        self.daily_rows_used = daily_rows_used

    @staticmethod
    def _next_reset(tz_name: str) -> datetime:
        """Calculate next quota reset (1st of next month)."""
        now = datetime.now(timezone.utc)
        if now.month == 12:
            next_month = datetime(now.year + 1, 1, 1, tzinfo=timezone.utc)
        else:
            next_month = datetime(now.year, now.month + 1, 1, tzinfo=timezone.utc)
        return next_month

    def get_limits(self) -> Dict[str, Any]:
        """Get limits for current tier."""
        return TIER_LIMITS.get(self.tier, TIER_LIMITS["free"])

    def check_can_run(
        self, estimated_rows: int = 0, credits_needed: int = 0
    ) -> Dict[str, Any]:
        """
        Check if a run is allowed based on monthly quotas.
        Returns: {allowed: bool, used_grace: bool, message: str}
        """
        limits = self.get_limits()

        # Check if quota reset is needed
        if datetime.now(timezone.utc) >= self.reset_at:
            return {
                "allowed": True,
                "used_grace": False,
                "needs_reset": True,
                "message": "Quota reset due",
            }

        # Check monthly rows quota
        if estimated_rows > 0:
            total_rows_after = self.monthly_rows_used + estimated_rows
            monthly_limit = limits["monthly_rows"]

            if total_rows_after > monthly_limit:
                # Check grace band
                grace_limit = monthly_limit * (1 + limits["grace_percentage"])
                if total_rows_after <= grace_limit and self.grace_uses_remaining > 0:
                    return {
                        "allowed": True,
                        "used_grace": True,
                        "message": f"Using grace allowance. {total_rows_after:,}/{int(grace_limit):,} rows this month. Upgrade for more.",
                        "show_upgrade": True,
                    }
                else:
                    remaining = max(0, monthly_limit - self.monthly_rows_used)
                    next_tier_limit = (
                        TIER_LIMITS["pro"]["monthly_rows"]
                        if self.tier == "free"
                        else TIER_LIMITS["scale"]["monthly_rows"]
                    )
                    return {
                        "allowed": False,
                        "message": f"Monthly row limit reached ({self.monthly_rows_used:,}/{monthly_limit:,}). You have {remaining:,} rows remaining. Resets {self.reset_at.strftime('%b %d')}. Upgrade for {next_tier_limit:,} rows/month.",
                        "show_upgrade": True,
                        "remaining": remaining,
                    }

        # Check enrichment credits
        if credits_needed > 0:
            total_credits_after = self.enrichment_credits_used + credits_needed
            credits_limit = limits["enrichment_credits"]

            if total_credits_after > credits_limit:
                remaining = max(0, credits_limit - self.enrichment_credits_used)
                return {
                    "allowed": False,
                    "message": f"Enrichment credit limit reached ({self.enrichment_credits_used}/{credits_limit}). Buy credit packs or upgrade.",
                    "show_upgrade": True,
                    "remaining_credits": remaining,
                }

        return {"allowed": True, "used_grace": False, "message": "Within quota limits"}

    def record_run(
        self, rows_returned: int, credits_consumed: int = 0, used_grace: bool = False
    ) -> None:
        """Record a completed run with row and credit usage."""
        self.monthly_rows_used += rows_returned
        self.enrichment_credits_used += credits_consumed
        # Legacy tracking
        self.daily_runs_used += 1
        self.daily_rows_used += rows_returned

        if used_grace:
            self.grace_uses_remaining = max(0, self.grace_uses_remaining - 1)
            self.last_grace_at = datetime.now(timezone.utc)

    def reset(self) -> None:
        """Reset monthly counters."""
        self.monthly_rows_used = 0
        self.enrichment_credits_used = 0
        # Legacy fields
        self.daily_runs_used = 0
        self.daily_rows_used = 0
        self.grace_uses_remaining = 1  # Reset grace
        self.reset_at = self._next_reset(self.timezone)


async def get_quota(account_id: str, db: AsyncSession) -> UsageQuota:
    """Get or create quota for account."""
    from ..models import UsageQuotaModel

    result = await db.execute(
        select(UsageQuotaModel).where(UsageQuotaModel.account_id == account_id)
    )
    quota_model = result.scalar_one_or_none()

    if not quota_model:
        # Create new quota
        quota_model = UsageQuotaModel(
            account_id=account_id,
            tier="free",
            monthly_rows_used=0,
            enrichment_credits_used=0,
            daily_runs_used=0,  # Legacy
            daily_rows_used=0,  # Legacy
            active_schedules=0,
            grace_uses_remaining=1,
            reset_at=UsageQuota._next_reset("UTC"),
            created_at=datetime.now(timezone.utc),
            updated_at=datetime.now(timezone.utc),
        )
        db.add(quota_model)
        await db.commit()

    # Handle models that don't have new fields yet (backwards compatibility)
    monthly_rows = getattr(
        quota_model, "monthly_rows_used", quota_model.daily_rows_used
    )
    enrichment_credits = getattr(quota_model, "enrichment_credits_used", 0)

    quota = UsageQuota(
        account_id=quota_model.account_id,
        tier=quota_model.tier,
        monthly_rows_used=monthly_rows,
        enrichment_credits_used=enrichment_credits,
        active_schedules=quota_model.active_schedules,
        grace_uses_remaining=quota_model.grace_uses_remaining,
        last_grace_at=quota_model.last_grace_at,
        reset_at=quota_model.reset_at if (quota_model.reset_at is None or quota_model.reset_at.tzinfo is not None) else quota_model.reset_at.replace(tzinfo=timezone.utc),
        timezone=getattr(quota_model, "timezone", "UTC"),
        # Legacy compatibility
        daily_runs_used=quota_model.daily_runs_used,
        daily_rows_used=quota_model.daily_rows_used,
    )

    # Auto-reset if needed
    if datetime.now(timezone.utc) >= quota.reset_at:
        quota.reset()
        update_values = {
            "monthly_rows_used": 0,
            "enrichment_credits_used": 0,
            "daily_runs_used": 0,
            "daily_rows_used": 0,
            "grace_uses_remaining": 1,
            "reset_at": quota.reset_at,
            "updated_at": datetime.now(timezone.utc),
        }
        await db.execute(
            update(UsageQuotaModel)
            .where(UsageQuotaModel.account_id == account_id)
            .values(**update_values)
        )
        await db.commit()

    return quota


async def record_quota_usage(
    account_id: str,
    db: AsyncSession,
    rows_returned: int,
    credits_consumed: int = 0,
    used_grace: bool = False,
) -> None:
    """Record quota usage after successful run."""
    from ..models import UsageQuotaModel

    update_values = {
        "monthly_rows_used": UsageQuotaModel.monthly_rows_used + rows_returned,
        "enrichment_credits_used": UsageQuotaModel.enrichment_credits_used
        + credits_consumed,
        # Legacy fields
        "daily_runs_used": UsageQuotaModel.daily_runs_used + 1,
        "daily_rows_used": UsageQuotaModel.daily_rows_used + rows_returned,
        "updated_at": datetime.now(timezone.utc),
    }

    if used_grace:
        update_values["grace_uses_remaining"] = UsageQuotaModel.grace_uses_remaining - 1
        update_values["last_grace_at"] = datetime.now(timezone.utc)

    await db.execute(
        update(UsageQuotaModel)
        .where(UsageQuotaModel.account_id == account_id)
        .values(**update_values)
    )
    await db.commit()



from g_gremlin.lint import (
    check_error_values,
    check_iferror_args,
    check_offset_args,
    check_unquoted_sheet_refs,
)


def test_unquoted_sheet_ref_with_space():
    issues = check_unquoted_sheet_refs("=A1/My Sheet!B10")
    assert len(issues) == 1
    assert issues[0]["type"] == "unquoted_sheet_ref"
    assert issues[0]["suggested_fix"] == "=A1/'My Sheet'!B10"


def test_unquoted_sheet_ref_with_hyphen():
    issues = check_unquoted_sheet_refs("=A1/FY-26!B10")
    assert len(issues) == 1
    assert issues[0]["suggested_fix"] == "=A1/'FY-26'!B10"


def test_unquoted_sheet_ref_with_underscore_ok():
    issues = check_unquoted_sheet_refs("=A1/My_Sheet!B10")
    assert len(issues) == 0


def test_offset_too_few_args():
    issues = check_offset_args("=OFFSET(A1, 1)")
    assert len(issues) == 1
    assert issues[0]["args_found"] == 2


def test_offset_correct_args():
    issues = check_offset_args("=OFFSET(A1, 0, -2, 1, 1)")
    assert len(issues) == 0


def test_iferror_missing_fallback():
    issues = check_iferror_args("=IFERROR(OFFSET(A1,0,-2))")
    assert len(issues) == 1


def test_iferror_with_fallback():
    issues = check_iferror_args("=IFERROR(OFFSET(A1,0,-2), 0)")
    assert len(issues) == 0


def test_error_values():
    assert check_error_values("#REF!")["type"] == "ref_error"
    assert check_error_values("#NAME?")["type"] == "name_error"
    assert check_error_values(123) is None



"""Periodic reservation sweeper entrypoint."""

from __future__ import annotations

import asyncio
import logging
from typing import Any, Dict

from ...db import AsyncSessionLocal
from ..enrichment_usage_tracker import EnrichmentUsageTracker

logger = logging.getLogger(__name__)


async def sweep_once() -> Dict[str, Any]:
    """Expire abandoned reservations and purge stale idempotency keys."""
    async with AsyncSessionLocal() as session:
        tracker = EnrichmentUsageTracker(session)
        summary = await tracker.sweep_timeouts()
        logger.info("reservation_sweeper_summary %s", summary)
        return summary


def sweep_sync() -> Dict[str, Any]:
    """Synchronous helper for cron/management commands."""
    return asyncio.run(sweep_once())



"""Shared utilities for renewal-based scouts."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, List, Tuple

from .base import BaseScoutCheck, PreflightResult
from ..crm_provider import Filter, FilterExpr, FilterOp
from ..utils import parse_sf_datetime


@dataclass
class RenewalConfig:
    object_type: str
    renewal_date_field: str
    account_id_field: str
    account_name_field: str


class BaseRenewalCheck(BaseScoutCheck):
    """Shared renewal field resolution and preflight checks."""

    def _resolve_renewal_config(self) -> RenewalConfig:
        crm = (self.provider.name or "salesforce").lower()
        if crm == "hubspot":
            return RenewalConfig(
                object_type="account",
                renewal_date_field="RenewalDate",
                account_id_field="AccountId",
                account_name_field="AccountName",
            )
        if crm == "dynamics":
            return RenewalConfig(
                object_type="contract",
                renewal_date_field="RenewalDate",
                account_id_field="AccountId",
                account_name_field="AccountName",
            )
        return RenewalConfig(
            object_type="contract",
            renewal_date_field="RenewalDate",
            account_id_field="AccountId",
            account_name_field="AccountName",
        )

    def _mapped_field(self, object_type: str, canonical_field: str) -> str:
        crm_key = (self.provider.name or "salesforce").lower()
        overrides = self.definition.field_overrides or {}
        crm_overrides = overrides.get(crm_key) or overrides.get(crm_key.lower()) or {}
        obj_key = object_type.lower()
        obj_overrides = crm_overrides.get(obj_key) or crm_overrides.get(object_type) or {}
        # Allow case-insensitive field keys
        for key in (canonical_field, canonical_field.lower()):
            if key in obj_overrides:
                return str(obj_overrides[key])
        return canonical_field

    def _available_fields(self, object_type: str) -> set[str]:
        try:
            desc = self.provider.describe(object_type)
        except Exception:
            return set()
        if isinstance(desc, dict):
            if "result" in desc:
                fields = desc.get("result", {}).get("fields", []) or []
                names = {str(f.get("name", "")) for f in fields if isinstance(f, dict)}
                return {n for n in names if n}
            if "properties" in desc:
                props = desc.get("properties", []) or []
                names = {str(p.get("name", "")) for p in props if isinstance(p, dict)}
                return {n for n in names if n}
        return set()

    def _filter_fields(self, object_type: str, fields: List[str]) -> List[str]:
        available = {f.lower() for f in self._available_fields(object_type)}
        if not available:
            return fields
        kept: List[str] = []
        for field in fields:
            mapped = self._mapped_field(object_type, field)
            if "." in mapped:
                kept.append(field)
                continue
            if mapped.startswith("_") and mapped.endswith("_value"):
                kept.append(field)
                continue
            if mapped.lower() in available:
                kept.append(field)
            else:
                self._warnings.append(f"Field '{mapped}' not found on {object_type}; skipping.")
        return kept

    def preflight(self) -> PreflightResult:
        cfg = self._resolve_renewal_config()
        if not self.provider.supports_object(cfg.object_type):
            return PreflightResult(ok=False, unsupported=True, reason=f"Unsupported object: {cfg.object_type}")

        mapped_date = self._mapped_field(cfg.object_type, cfg.renewal_date_field)
        available = {f.lower() for f in self._available_fields(cfg.object_type)}
        if available and mapped_date.lower() not in available:
            return PreflightResult(
                ok=False,
                unsupported=True,
                reason=(
                    f"Field '{mapped_date}' not found on {cfg.object_type}. "
                    "Map RenewalDate via field_overrides in scout YAML."
                ),
            )
        return PreflightResult(ok=True)

    def _today_values(self, horizon_days: int | None = None) -> tuple[Any, Any | None]:
        now = datetime.now(timezone.utc)
        today = datetime(year=now.year, month=now.month, day=now.day, tzinfo=timezone.utc)
        if self.provider.name == "hubspot":
            start = int(today.timestamp() * 1000)
            if horizon_days is None:
                return start, None
            end = int((today + timedelta(days=horizon_days)).timestamp() * 1000)
            return start, end
        if self.provider.name == "dynamics":
            if horizon_days is None:
                return today, None
            return today, today + timedelta(days=horizon_days)
        start = today.date().isoformat()
        if horizon_days is None:
            return start, None
        end = (today.date() + timedelta(days=horizon_days)).isoformat()
        return start, end

    def _days_until(self, value: str | None) -> int | None:
        dt = parse_sf_datetime(value)
        if not dt:
            return None
        now = datetime.now(timezone.utc)
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=timezone.utc)
        return (dt - now).days

    def _days_overdue(self, value: str | None) -> int | None:
        dt = parse_sf_datetime(value)
        if not dt:
            return None
        now = datetime.now(timezone.utc)
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=timezone.utc)
        return (now - dt).days


__all__ = ["BaseRenewalCheck", "RenewalConfig"]
